from zpy.constants import EnvironmentKeys
from typing import Dict, List
from zpy.utils import find
from zpy.utils import env as genv
import os

__author__ = "Noé Cruz | contactozurckz@gmail.com"
__copyright__ = "Copyright 2021, Small APi Project"
__credits__ = ["Noé Cruz", "Zurck'z"]
__license__ = "MIT"
__version__ = "0.0.1"
__maintainer__ = "Noé Cruz"
__email__ = "contactozurckz@gmail.com"
__status__ = "Dev"


def get_schema(fn: str, env=None, schemas=None) -> str:
    """
    Get current schema according environment

    PD: You should be configure schemas list in repositories
    """
    real_env = os.getenv("env") if env is None else env
    if real_env is None:
        raise "Environment not configured"
    schema = find(schemas, lambda sc: sc["env"] == real_env)
    if schema:
        return "{}.{}".format(schema["value"], fn)
    raise "Not schema configured"


def get_schema_by(schemas: List[Dict], env=None) -> str:
    """
    Get current schema according environment

    PD: You should be configure schemas list in repositories
    """
    real_env = os.getenv("env") if env is None else env
    if real_env is None:
        raise None
    schema = find(schemas, lambda sc: sc["env"] == real_env)
    if schema:
        return schema["value"]
    return None


def get_current_schema(schemas: List[Dict], local_emv: str = None, global_env: str = None):
    real_env = global_env if local_emv is None else local_emv
    external_env = genv(str(EnvironmentKeys.ENVIRONMENT))
    real_env = external_env if real_env is None else real_env
    if real_env is not None:
        if schemas is not None:
            db_schema = get_schema_by(schemas, real_env)
            return db_schema
    return None


from abc import ABC, abstractmethod
from typing import Any, List

from zpy.custom.models import DBSSMCredentialsModel, SSMConstants
from zpy.cloud.aws.ssm import SSMParameter
from zpy.api import ApiConfiguration
from zpy.db import DBConnection
from zpy.db.oracle import OracleDB
import logging
import os


class Plugin(ABC):
    @abstractmethod
    def initialize(self, config: ApiConfiguration, ssm: SSMParameter, *args):
        pass


# Receive this by parameter where use
# class TradeConstants(Enum):
#     SSM_PREFIX = "/aws/reference/secretsmanager/com/upax/trade"
#     LOCAL_DB_SSM = "/db/oracle/zeuststtrade"
#     DB_SSM = "/db/oracle/usrtrade"
#     AES_SK_SSM = "/security/encryption"
#     AWS_SSM = "/aws/services"


# Database Schemas for Trade Project
DB_SCHEMES = [
    {"env": "prod", "value": "TRADE"},
    {"env": "qa", "value": "TRADE"},
    {"env": "dev", "value": "TRADE"},
    {"env": "localdev", "value": "ZEUSTSTTRADE"},
]

# Response modifier acordin zcommons procesor

custom_response_mapper = lambda r: {"message": r}


def pool_initialize(config: ApiConfiguration):
    """
    Initialization of the connection db connection pool and
    local client in case we are in local dev environment.
    """
    db: DBConnection = OracleDB(
        config=config.to_dict(), schemas=DB_SCHEMES, env=config.ENVIRONMENT
    )

    try:
        if config.ENVIRONMENT == "localdev":
            db.init_local_client(os.getenv("LOCAL_CLIENT"))
        db.initialize_pool(
            dns=db.get_tsn_dsn_conenction(config.to_dict(), "DSN"),
            homogeneous=True,
            user=config.DB_USER,
            pwd=config.DB_PASSWORD,
        )
    except Exception as e:
        print(e)
    return db


def setup_env(vars: dict):
    for k in vars.keys():
        os.environ[k] = vars[k]


def load_plugins(
    plugins: List[Plugin] = [],
    args_plugs: List[List[Any]] = [],
    env: str = None,
    ssm_const: SSMConstants = None,
):
    try:
        environment = os.getenv("env") if env == None else env
        if environment == None:
            raise Exception(
                "Environment not found. set environment variable with name: 'env'  and value: 'prod|qa|dev|localdev' "
            )
        if ssm_const == None:
            raise Exception("SSM Constants cannot be null, set ssm constants: see: ")
        ssm = SSMParameter(with_profile=True, prefix=ssm_const.smm_prefix)
        db_prefix = (
            ssm_const.local_db_ssm if environment == "localdev" else ssm_const.db_smm
        )
        db_credentials: DBSSMCredentialsModel = ssm.get(prefix=db_prefix, model=DBSSMCredentialsModel)
        # setup_env(db_credentials)
        aes_credentials = ssm.get(prefix=ssm_const.aes_sk)
        # setup_env(aes_credentials)
        config: ApiConfiguration = ApiConfiguration()
        config.DB_URI = db_credentials.uri
        config.DB_PASSWORD = db_credentials.password
        config.DB_PORT = db_credentials.port
        config.DB_SERVICE = db_credentials.service
        config.DB_USER = db_credentials.user
        config.ENVIRONMENT = os.getenv("env")
        config.SECRET = os.getenv("API_SECRET")
        config.CUSTOM = aes_credentials
        plugins_result = []
        for i, plugin in enumerate(plugins):
            args = args_plugs[i] if i < len(args_plugs) else {}
            plugins_result.append(plugin().initialize(config=config, ssm=ssm, **args))
        return (config, plugins_result)
    except Exception as e:
        logging.exception(e)


def setup_microservice(
    plugins: List[Plugin] = [],
    args_plugs: List[List[Any]] = [],
    env: str = None,
    ssm_const: SSMConstants = None
):
    """

    Deprecated function, will be remove in 1.0.0 core version.
    Use instead load_plugins function.

    """
    try:
        environment = os.getenv("env") if env == None else env
        if environment == None:
            raise Exception(
                "Environment not found. set environment variable with name: 'env'  and value: 'prod|qa|dev|localdev' "
            )
        if ssm_const == None:
            raise Exception("SSM Constants cannot be null, set ssm constants: see: ")
        ssm = SSMParameter(with_profile=True, prefix=ssm_const.smm_prefix)
        db_prefix = (
            ssm_const.local_db_ssm
            if environment == "localdev"
            else ssm_const.db_smm
        )
        db_credentials: DBSSMCredentialsModel = ssm.get(prefix=db_prefix, model=DBSSMCredentialsModel)
        # setup_env(db_credentials)
        aes_credentials = ssm.get(prefix=ssm_const.aes_sk)
        setup_env(aes_credentials)
        config: ApiConfiguration = ApiConfiguration()
        config.DB_URI = db_credentials.uri
        config.DB_PASSWORD = db_credentials.password
        config.DB_PORT = db_credentials.port
        config.DB_SERVICE = db_credentials.service
        config.DB_USER = db_credentials.user
        config.ENVIRONMENT = os.getenv("env")
        config.SECRET = os.getenv("API_SECRET")
        db: DBConnection = pool_initialize(config)
        plugins_result = []
        for i, plugin in enumerate(plugins):
            plugins_result.append(
                plugin().initialize(config=config, ssm=ssm, *args_plugs[i])
            )

        return (config, db, plugins_result)
    except Exception as e:
        logging.exception(e)



from datetime import datetime
import uuid
import os

__author__ = "Noé Cruz | contactozurckz@gmail.com"
__copyright__ = "Copyright 2021, Small APi Project"
__credits__ = ["Noé Cruz", "Zurck'z"]
__license__ = "upax"
__version__ = "0.0.1"
__maintainer__ = "Noé Cruz"
__email__ = "contactozurckz@gmail.com"
__status__ = "Dev"

# Date utils
str_date = lambda str_date, format="%Y-%m-%d": datetime.strptime(str_date, format)
date_str = lambda date_time, format="%Y-%m-%d %H:%M:%S": date_time.strftime(format)
get_date = lambda: datetime.now()


def add_hours_to_date(date: str, time: str):
    """
    date: 2020-10-10
    time: 10:50:00
    """
    dte = str_date(date)
    splitted_time = time.split(r":")
    size = len(splitted_time)
    nwdte = None
    try:
        if size >= 3:
            nwdte = dte.replace(
                hour=int(splitted_time[0]),
                minute=int(splitted_time[1]),
                second=int(splitted_time[2]),
            )
        elif size == 2:
            nwdte = dte.replace(
                hour=int(splitted_time[0]), minute=int(splitted_time[1]), second=0
            )
        else:
            nwdte = dte.replace(hour=int(splitted_time[0]), minute=0, second=0)
    except Exception as e:
        nwdte = dte.replace(hour=0, minute=0, second=0)

    return nwdte


# Environment utils
def env(key: str):
    """
    Get environment value by key
    """
    try:
        return os.getenv(key)
    except Exception as e:
        return None


def find(list, filter):
    for x in list:
        if filter(x):
            return x
    return None


def get_operation_id() -> str:
    return str(uuid.uuid1())


from resources.greeting_resoource import GreetingResource
from config.api_configuration import apiConfig
from zpy.api import ZResource, api



@api(base="/api", config=apiConfig)
def create_api():
    return [ZResource("/greeting", GreetingResource())]



app = create_api()

# 🚨 Only use it in local tests 💻
if __name__ == "__main__":
    app.run(host="localhost", debug=True)



from typing import Any, Callable, List
import logging


def safely_exec(callable: Callable, args: List[Any] = []):
    try:
        return callable(*args)
    except Exception as e:
        logging.exception(e)
    return None


def exec_ifnt_null(callable: Callable, args: List[Any] = []):
    """
    Execute function if args not null
    """
    for arg in args:
        if arg == None:
            return False
    return callable(*args)


from zpy.constants import Messages
from typing import Any, Dict, List, Tuple, Union
from zpy.api.reponse import BResponse, Builder, Status
from zpy.api.stages import ResourceLayer
from zpy.utils import get_operation_id
from zpy.api.exceptions import ZError
from marshmallow_objects import models
from marshmallow.exceptions import ValidationError
from zpy.logger import TLogger, c_warn
from flask.views import MethodView
from flask import request
from enum import Enum
import json

__author__ = "Noé Cruz | contactozurckz@gmail.com"
__copyright__ = "Copyright 2021, Small APi Project"
__credits__ = ["Noé Cruz", "Zurck'z"]
__license__ = "MIT"
__version__ = "0.0.1"
__maintainer__ = "Noé Cruz"
__email__ = "contactozurckz@gmail.com"
__status__ = "Dev"


class HTTP_METHODS(Enum):
    GET = 1
    POST = 2
    PUT = 3
    PATCH = 4
    DELETE = 5


class ZResource(MethodView):
    __name__ = "ZResource"
    blocked_methods = []

    def __init__(self) -> None:
        super().__init__()
        print(f"Zurck'z  - Core Resource was iniitalized")

    def __method_not_allowed(self):
        return Builder().add_status(Status.METHOD_NOT_ALLOWED).build()

    def __is_not_allowed(self, method: HTTP_METHODS):
        return method in self.blocked_methods

    def get(self):
        if self.__is_not_allowed(HTTP_METHODS.GET):
            return self.__method_not_allowed()

    def post(self):
        if self.__is_not_allowed(HTTP_METHODS.POST):
            return self.__method_not_allowed()

    def put(self):
        if self.__is_not_allowed(HTTP_METHODS.PUT):
            return self.__method_not_allowed()

    def patch(self):
        if self.__is_not_allowed(HTTP_METHODS.PATCH):
            return self.__method_not_allowed()

    def delete(self):
        if self.__is_not_allowed(HTTP_METHODS.DELETE):
            return self.__method_not_allowed()

    def new_operation(self):
        """
        Create a new http operation.

        Returns:
            l Logger object
            i Operation id
        """
        operation_id = get_operation_id()
        log = TLogger(request.path, operation_id)
        log.add_phase(phase=ResourceLayer())
        if "request" in request.environ:
            log.add_info("Request:")
            log.add_info(request.environ["request"])
        return log, operation_id

    def close_operation(
        self, l: TLogger = None, e: Exception = None, failed: bool = False
    ):
        """
        Dispose current logger.

        l: TLogger
            Logger object
        e: Exception
            Exception to add into logger stream
        failed: bool
            Determine if logger stream should be printed
        """
        if l != None:
            if failed:
                l.add_exception(e)
                l.show_stack_trace()
            l.dispose()

    def get_request(self) -> dict:
        """
        Get current request

        Raises
        ------
        NotValueProvided
            If no values is set for the response or passed in as a parameter.

        Returns
        -------
        request : Dict[Any]
            Dict with request content

        """
        try:
            if "request" in request.environ and request.environ["request"] != None:
                return json.loads(request.environ["request"])
            return {}
        except Exception as e:
            return {}

    def either_validation(
        self, request: dict, model: models.Model
    ) -> Tuple[models.Model, Dict]:
        """
        Validate request acordning model specification

        Parameters
        ----------
        request : dict | required
            Request object to validate

        model : marshmallow_objects.models.Model | required
            The model specification

        Raises
        ------
        NotValueProvided
            If no values is set for the response or passed in as a parameter.

        Returns
        -------
        (model,errors) : Tuple[models.Model, Dict]
            Tuple with model and erros either the validation process

        """

        model_result: Union[List, models.Model] = None
        errors: Union[List, None] = None
        try:
            if request == None or len(request.items()) == 0:
                return None, {
                    "request": "The request was not provided, validation request error",
                    "fields": f"{model().__missing_fields__} not provided",
                }
            model_result = model(**request)
        except ValidationError as e:
            model_result = e.valid_data
            if isinstance(e.messages, Dict):
                errors = [e.messages]
            else:
                errors = e.messages
        return model_result, errors

    def success(
        self,
        payload={},
        logger: TLogger = None,
        status_code: Status = Status.SUCCESS,
        operation_id=None,
        description: str = None,
        map=None,
    ):
        """
        Build success response and dispose logger
        """
        self.close_operation(logger, None)
        return BResponse.success(payload, status_code, operation_id, description, map)

    def bad_request(
        self,
        payload=None,
        logger: TLogger = None,
        errors: List[Any] = [],
        status_code: Status = Status.BAD_REQUEST,
        operation_id=None,
        description: str = None,
        map=None,
    ):
        """
        Build bad request response and dispose logger
        """
        self.close_operation(logger, None)
        return BResponse.bad_request(
            payload, errors, status_code, operation_id, description, map
        )

    def response(
        self,
        payload={},
        logger: TLogger = None,
        errors: List[Any] = [],
        status_code: Status = Status.SUCCESS,
        operation_id=None,
        description: str = None,
        map=None,
    ):
        """
        Build any response and dispose logger
        """
        self.close_operation(logger, None)
        return BResponse.bad_request(
            payload, errors, status_code, operation_id, description, map
        )

    def handlde_exceptions(
        self,
        exception: Union[ZError, Exception],
        logger: TLogger,
        operation_id: str,
        payload=None,
        custom_errors: List[Union[str, dict]] = None,
        custom_message: str = None,
        custom_status: Status = None,
        map=None,
    ) -> Tuple[Dict, int]:
        """
        Handle exception and build error response with default or custom data

        Parameters
        ----------
        exception : Union[Exception,ZError]
            Exception object to handler

        logger : ZLogger
            Operation Logger for close operation and show trace

        operation_id: str
            Operation ID, must be the same as the initial ID of the operation

        custom_errors: Optional[List[Union[str, dict]]]
            Errors list for add in the resopnse
        custom_message: str
            Custom response description
        custom_statis: Status
            Cutom status for response
        map: Fn Map, Optional
            Response Modifier function
        Raises
        ------
        NotValueProvided
            If no values is set for the response or passed in as a parameter.

        Returns
        -------
        HttpResponse : Tuple[Dict,int]
            Response with error code or custom data

        """
        c_warn("handlde_exceptions() will be deprecated in version 1.0.0, use handle_exceptions instead.")
        return self.handle_exceptions(
            exception,
            logger,
            operation_id,
            payload,
            custom_errors,
            custom_message,
            custom_status,
            map,
        )

    def handle_exceptions(
        self,
        exception: Union[ZError, Exception],
        logger: TLogger,
        operation_id: str,
        payload=None,
        custom_errors: List[Union[str, dict]] = None,
        custom_message: str = None,
        custom_status: Status = None,
        map=None,
    ) -> Tuple[Dict, int]:
        """    
        Handlde exception and build error response with default or custom data

        Parameters
        ----------
        exception : Union[Exception,ZError]
            Exception object to handler

        logger : ZLogger
            Operation Logger for close operation and show trace

        operation_id: str
            Operation ID, must be the same as the initial ID of the operation

        custom_errors: Optional[List[Union[str, dict]]]
            Errors list for add in the resopnse
        custom_message: str
            Custom response description
        custom_statis: Status
            Cutom status for response
        map: Fn Map, Optional
            Response Modifier function
        Raises
        ------
        NotValueProvided
            If no values is set for the response or passed in as a parameter.

        Returns
        -------
        HttpResponse : Tuple[Dict,int]
            Response with error code or custom data

        """
        errors = [] if custom_errors == None else custom_errors
        final_status = (
            Status.INTERNAL_SERVER_ERROR if custom_status == None else custom_status
        )
        if isinstance(exception, ZError):
            errors.append(exception.get_error())
            logger.add_info(exception.get_message())
            self.close_operation(logger, exception, True)
            return BResponse.err(
                errors=errors,
                operation_id=operation_id,
                description=custom_message,
                status=final_status,
                payload=payload,
                fn_transform=map,
            )
        else:
            errors.append(str(Messages.GEN_MSG_ERROR))
            logger.add_info(str(Messages.GEN_MSG_ERROR))
            self.close_operation(logger, exception, True)
            return BResponse.err(
                errors=errors,
                operation_id=operation_id,
                description=custom_message,
                status=final_status,
                payload=payload,
                fn_transform=map,
            )



import gzip
import json

__author__ = "Noé Cruz | contactozurckz@gmail.com"
__copyright__ = "Copyright 2007, The Cogent Project"
__credits__ = ["Noé Cruz", "Zurck'z", "Jesus Salazar"]
__license__ = "MIT"
__version__ = "0.0.1"
__maintainer__ = "Noé Cruz"
__email__ = "contactozurckz@gmail.com"
__status__ = "Dev"

##  Class GZip
#   Zipper class contains compress/decompress gzip methods
#
class GZip():
    """
    Helper compress/decompress gzip
    """

    @staticmethod
    def unzip_json( json_data_compressed) -> dict:
        """Decompress json compressed provided

        Parameters
        ----------
        data : bytes | str
            JSON data to descompress

        Returns
        -------
        dict
            JSON descompressed
        """
        return json.loads(GZip.decompress(json_data_compressed))

    @staticmethod
    def gzip_json( json_data: dict) -> bytes:
        """Compress JSON with gzip

        Parameters
        ----------
        data : str
            JSON data to compress
        enc : str, optional
            charset encoding (default is
            utf-8)

        Returns
        -------
        bytes
            Data compressed
        """
        return GZip.compress(json.dumps(json_data))

    @staticmethod
    def compress( data: str, enc = 'utf-8') -> bytes:
        """Decompress data provided

        Parameters
        ----------
        data : str
            Data to compress
        enc : str, optional
            charset encoding (default is
            utf-8)

        Returns
        -------
        bytes
            Data compressed
        """
        return gzip.compress(bytes(data,enc))

    @staticmethod
    def decompress(data: bytes, enc = 'utf-8') -> str:
        """Decompress data provided

        Parameters
        ----------
        data : str
            Gzip compressed data
        enc : str, optional
            charset encoding (default is
            utf-8)

        Returns
        -------
        str
            Data descompressed in string format
        """
        return gzip.decompress(data).decode(enc)


from abc import ABC, abstractmethod
from flask import request
from zpy.logger import g_log
from zpy.api.errors import ErrorBuilder
from flask import Flask
from zpy.utils.Encryptor import AESEncryptor
from zpy.api.stages import Encryption, ResourceGateway
from zpy.api.reponse import Builder, Status
import json

__author__ = "Noé Cruz | contactozurckz@gmail.com"
__copyright__ = "Copyright 2021, Small APi Project"
__credits__ = ["Noé Cruz", "Zurck'z"]
__license__ = "upax"
__version__ = "0.0.1"
__maintainer__ = "Noé Cruz"
__email__ = "contactozurckz@gmail.com"
__status__ = "Dev"


class ZHook(ABC):
    @abstractmethod
    def execute(self, app: Flask, **kwargs):
        pass


# Custom hooks


class AESEncryptHook(ZHook):
    def execute(self, app, **kwargs):
        @app.after_request
        def encrypt(response: Flask):
            try:
                if response.data:
                    parsed = json.loads(response.data)
                    aes: str = None
                    if (
                        kwargs != None and "aes_sk" in kwargs
                    ):  # ! WARNING HARD KEY FOR EXTARCT AES SK
                        aes = kwargs["aes_sk"]
                    encrypted = AESEncryptor.encrypt_ws_request(parsed, secret_key=aes)
                    response.data = json.dumps(encrypted)
                return response
            except Exception as e:
                stage = Encryption()
                g_log(e, stage)
                return Builder().error(
                    {},
                    errors=[
                        ErrorBuilder.common(
                            "Encryption hook throw exception",
                            "The response catched by hook have invalid structure",
                            stage,
                        )
                    ],
                )

        return super().execute(app)


class AESDecryptHook(ZHook):
    def execute(self, app: Flask, **kwargs):
        @app.before_request
        def decrypt():
            try:
                if request.data:
                    aes: str = None
                    if (
                        kwargs != None and "aes_sk" in kwargs
                    ):  # ! WARNING HARD KEY FOR EXTARCT AES SK
                        aes = kwargs["aes_sk"]
                    decrypt_data = AESEncryptor.decrypt_ws_response(
                        request.get_json(), aes
                    )
                    request.data = json.dumps(decrypt_data)
                return request
            except Exception as e:
                stage = Encryption()
                g_log(e, stage)
                return Builder().error(
                    {},
                    errors=[
                        ErrorBuilder.common(
                            "Encryption hook throw exception",
                            "The response catched by hook have invalid structure",
                            stage,
                        )
                    ],
                )

        return super().execute(app)


class NotFoundHook(ZHook):
    def execute(self, app: Flask, **kwargs):
        @app.errorhandler(404)
        def not_found(e):
            try:
                g_log(e, ResourceGateway())
                return (
                    Builder()
                    .add_status(Status.NOT_FOUND)
                    .add_error("Resource not exist")
                    .add_error(str(e))
                    .build()
                )
            except Exception:
                pass

        return super().execute(app)


class TearDownHook(ZHook):
    def execute(self, app: Flask, **kwargs):
        @app.teardown_appcontext
        def dispose(exception):
            try:
                print("Teardown Micro Service")
            except Exception:
                pass

        return super().execute(app)



from zpy.api.errors import ErrorBuilder

__author__ = "Noé Cruz | contactozurckz@gmail.com"
__copyright__ = "Copyright 2021, Small APi Project"
__credits__ = ["Noé Cruz", "Zurck'z"]
__license__ = "upax"
__version__ = "0.0.1"
__maintainer__ = "Noé Cruz"
__email__ = "contactozurckz@gmail.com"
__status__ = "Dev"


class ZError(Exception):
    """
    ZPy Base Error
    """

    CUSTOM_MESSAGE: str = (
        "An uncontrolled, developer-unspecified error has been generated"
    )
    custom_message: str = None
    custom_reason: str = None
    exception_data: Exception = None

    def get_error(self):
        return ErrorBuilder.common(
            reason= self.custom_reason,
            message= self.get_message(),
        )

    def get_message(self):
        return (
            self.CUSTOM_MESSAGE if self.custom_message == None else self.custom_message
        )

    def get_exception(self):
        if self.exception_data == None:
            msg = (
                self.CUSTOM_MESSAGE
                if self.custom_message == None
                else self.custom_message
            )
            return Exception(msg)
        return self.exception_data

    def __init__(
        self,
        message: str = None,
        d_except: Exception = None,
        reason: str = None,
        *args: object
    ) -> None:
        super().__init__(*args)
        self.custom_message = message
        self.exception_data = d_except
        self.custom_reason = reason


class ZRepositoryError(ZError):
    """
    General Repository Layer Error
    """

    CUSTOM_MESSAGE = "[RP-00] An error was generated in the data access layer when performing operations with the provided data."

    def __init__(
        self,
        message: str = None,
        d_except: Exception = None,
        reason: str = None,
        *args: object
    ) -> None:
        super().__init__(message=message, d_except=d_except, reason=reason, *args)

    """
    General Service Layer Error
    """

    CUSTOM_MESSAGE = (
        "[SE-00] An error was generated when try execute logics with data supplied"
    )

    def __init__(
        self,
        message: str = None,
        d_except: Exception = None,
        reason: str = None,
        *args: object
    ) -> None:
        super().__init__(message=message, d_except=d_except, reason=reason, *args)


class ZUnavailableError(ZError):
    """
    Unavaible Error
    """

    CUSTOM_MESSAGE = "[UE-00] An error was generated when trying to interact with some entity and it is not available"

    def __init__(
        self,
        message: str = None,
        d_except: Exception = None,
        reason: str = None,
        *args: object
    ) -> None:
        super().__init__(message=message, d_except=d_except, reason=reason, *args)


class ZDatabaseOperationError(ZError):
    """
    Database error
    """

    CUSTOM_MESSAGE = "[RP-DB-00] An error was generated when interacting with the data provider when performing an operation on the schema with the data provided"

    def __init__(
        self,
        message: str = None,
        d_except: Exception = None,
        reason: str = None,
        *args: object
    ) -> None:
        super().__init__(message=message, d_except=d_except, reason=reason, *args)


class ZDBIntegrityError(ZError):
    """
    * Integrity database violated error
    """

    CUSTOM_MESSAGE = "[RP-DB-01] An error was generated in the data provider, the integrity of the schema was violated by some operation carried out with the data provided"

    def __init__(
        self,
        message: str = None,
        d_except: Exception = None,
        reason: str = None,
        *args: object
    ) -> None:
        super().__init__(message=message, d_except=d_except, reason=reason, *args)


class ZResourceError(ZError):
    """
    Resource exception
    """

    CUSTOM_MESSAGE = "[RS-00] An error was generated on the resource layer when it try execute some operation [Validation, Request Extraction, Response Building...]"

    def __init__(
        self,
        message: str = None,
        d_except: Exception = None,
        reason: str = None,
        *args: object
    ) -> None:
        super().__init__(message=message, d_except=d_except, reason=reason, *args)


# Copyright (c) 2018 Bao Nguyen <b@nqbao.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

import boto3
from botocore.exceptions import ClientError
import datetime


class SSMParameterStore(object):
    """
    Provide a dictionary-like interface to access AWS SSM Parameter Store
    """

    def __init__(self, prefix=None, ssm_client=None, ttl=None):
        self._prefix = (prefix or "").rstrip("/") + "/"
        self._client = ssm_client or boto3.client("ssm")
        self._keys = None
        self._substores = {}
        self._ttl = ttl

    def get(self, name, **kwargs):
        assert name, "Name can not be empty"
        if self._keys is None:
            self.refresh()

        abs_key = "%s%s" % (self._prefix, name)
        if name not in self._keys:
            if "default" in kwargs:
                return kwargs["default"]

            raise KeyError(name)
        elif self._keys[name]["type"] == "prefix":
            if abs_key not in self._substores:
                store = self.__class__(
                    prefix=abs_key, ssm_client=self._client, ttl=self._ttl
                )
                store._keys = self._keys[name]["children"]
                self._substores[abs_key] = store

            return self._substores[abs_key]
        else:
            return self._get_value(name, abs_key)

    def refresh(self):
        self._keys = {}
        self._substores = {}

        paginator = self._client.get_paginator("describe_parameters")
        pager = paginator.paginate(
            ParameterFilters=[
                dict(Key="Path", Option="Recursive", Values=[self._prefix])
            ]
        )

        for page in pager:
            for p in page["Parameters"]:
                paths = p["Name"][len(self._prefix) :].split("/")
                self._update_keys(self._keys, paths)

    @classmethod
    def _update_keys(cls, keys, paths):
        name = paths[0]

        # this is a prefix
        if len(paths) > 1:
            if name not in keys:
                keys[name] = {"type": "prefix", "children": {}}

            cls._update_keys(keys[name]["children"], paths[1:])
        else:
            keys[name] = {"type": "parameter", "expire": None}

    def keys(self):
        if self._keys is None:
            self.refresh()

        return self._keys.keys()

    def _get_value(self, name, abs_key):
        entry = self._keys[name]

        # simple ttl
        if self._ttl == False or (
            entry["expire"] and entry["expire"] <= datetime.datetime.now()
        ):
            entry.pop("value", None)

        if "value" not in entry:
            parameter = self._client.get_parameter(Name=abs_key, WithDecryption=True)[
                "Parameter"
            ]
            value = parameter["Value"]
            if parameter["Type"] == "StringList":
                value = value.split(",")

            entry["value"] = value

            if self._ttl:
                entry["expire"] = datetime.datetime.now() + datetime.timedelta(
                    seconds=self._ttl
                )
            else:
                entry["expire"] = None

        return entry["value"]

    def __contains__(self, name):
        try:
            self.get(name)
            return True
        except:
            return False

    def __getitem__(self, name):
        return self.get(name)

    def __setitem__(self, key, value):
        raise NotImplementedError()

    def __delitem__(self, name):
        raise NotImplementedError()

    def __repr__(self):
        return "ParameterStore[%s]" % self._prefix



##  @packege zeus_security_py
#   Helper package for data security that will implement zeus microservices
#
#
from Cryptodome.Cipher import AES
from Cryptodome import Random
from hashlib import sha256
import base64
import os
import json

__author__ = "Noé Cruz | contactozurckz@gmail.com"
__copyright__ = "Copyright 2007, The Cogent Project"
__credits__ = ["Noé Cruz", "Zurck'z", "Jesus Salazar"]
__license__ = "MIT"
__version__ = "0.0.1"
__maintainer__ = "Noé Cruz"
__email__ = "contactozurckz@gmail.com"
__status__ = "Dev"

##  Class Encryptor
#   Encryptor class contains AES encrypt/decrypt functions
#
class AESEncryptor:
    """
    Helper class for data security this contains certain methods for it.

    AES (Advanced Encryption Standard) is a symmetric block cipher standardized by NIST .
    It has a fixed data block size of 16 bytes. Its keys can be 128, 192, or 256 bits long.

    Attributes
    ----------
    default_block_size : int
        Default block size for aes (default 32)
    _sk_env : str
        Key for get secret key from environment

    Methods
    -------
    __is_valid(sk=None)
        Check if the secret key of argument is null, if that is null try to get secret key from environment.
    encrypt
    """

    default_block_size: int = 32
    _sk_env = "AES_SK"

    @staticmethod
    def __is_valid(sk: str = None):
        if sk is not None:
            return sk
        sk_env: str = os.getenv(AESEncryptor._sk_env)
        if sk_env is not None:
            return sk_env
        raise Exception("AES Secret key was not provided!")

    @staticmethod
    def decrypt_ws_response(payload: dict, secret_key=None) -> dict:
        json_decrypted = AESEncryptor.decrypt(payload["data"], secret_key)
        return json_decrypted

    @staticmethod
    def encrypt_ws_request(payload: dict, secret_key=None) -> dict:
        encrypted_payload = AESEncryptor.encrypt(json.dumps(payload), secret_key)
        return {"data": encrypted_payload}

    @staticmethod
    def json_decrypt(json_encrypted: str, secret_key=None) -> dict:
        return json.loads(AESEncryptor.encrypt(json_encrypted, secret_key))

    @staticmethod
    def json_encrypt(json_to_encrypt: dict, secret_key=None) -> str:
        json_str = json.dumps(json_to_encrypt)
        return AESEncryptor.encrypt(json_str, secret_key)

    @staticmethod
    def json_decrypt(json_encrypted: str, secret_key=None) -> dict:
        return json.loads(AESEncryptor.encrypt(json_encrypted, secret_key))

    @staticmethod
    def encrypt(
        value: str,
        secret_key: str = None,
        aes_mode=AES.MODE_CBC,
        charset="utf-8",
        block_size: int = 16,
    ) -> str:
        secret_key = AESEncryptor.__is_valid(secret_key).encode(charset)
        raw_bytes = AESEncryptor.__pad(value)
        iv = Random.new().read(block_size)
        cipher = AES.new(secret_key, aes_mode, iv)
        return base64.b64encode(iv + cipher.encrypt(raw_bytes)).decode(charset)

    @staticmethod
    def decrypt(
        value: str, secret_key=None, aes_mode=AES.MODE_CBC, charset="utf-8"
    ) -> str:
        secret_key = str.encode(AESEncryptor.__is_valid(secret_key))
        encrypted = base64.b64decode(value)
        iv = encrypted[:16]
        cipher = AES.new(secret_key, aes_mode, iv)
        return AESEncryptor.__un_pad(cipher.decrypt(encrypted[16:])).decode(charset)

    @staticmethod
    def genHash(value: str, charset="utf-8") -> str:
        return sha256(value.encode(charset)).hexdigest()

    @staticmethod
    def __pad(s: str, block_size: int = 16, charset: str = "utf-8") -> bytes:
        return bytes(
            s
            + (block_size - len(s) % block_size)
            * chr(block_size - len(s) % block_size),
            charset,
        )

    @staticmethod
    def __un_pad(value: str) -> str:
        return value[0 : -ord(value[-1:])]



from typing import Dict, Union
from zpy.api.models import ZModel
from zpy.cloud.aws import AWSCredentials, AWS_DEFAULT_REGION
import boto3
import json


__author__ = "Noé Cruz | contactozurckz@gmail.com"
__copyright__ = "Copyright 2021, Small APi Project"
__credits__ = ["Noé Cruz", "Zurck'z"]
__license__ = "MIT"
__version__ = "0.0.1"
__maintainer__ = "Noé Cruz"
__email__ = "contactozurckz@gmail.com"
__status__ = "Dev"


class SSMParameter:

    ssm_client = None
    with_profile: bool = False
    store = {}
    prefix: str = None

    def __init__(
        self,
        credentials: AWSCredentials = None,
        with_profile: bool = True,
        prefix: str = None,
    ) -> None:
        self.with_profile = with_profile
        self.prefix = prefix
        if with_profile or credentials is None:
            self.ssm = boto3.client("ssm", region_name=AWS_DEFAULT_REGION)
        else:
            self.ssm = boto3.client(
                "ssm",
                aws_access_key_id=credentials.access_key,
                aws_secret_access_key=credentials.secret_key,
                region_name=credentials.region,
            )

    def get_from_cache(self, name: str, model: ZModel = None) -> Union[Dict, ZModel]:
        """
        Get parameter stored in cache.
        """
        if name not in self.store:
            return None
        data = self.store[name]
        if model == None:
            return data
        return model(**data)

    def get(
        self,
        prefix: str = None,
        decryption: bool = True,
        store: bool = False,
        store_name: str = "",
        model: ZModel = None,
        refresh: bool = False,
    ) -> Union[Dict, ZModel]:
        """
        Get paramater from AWS SSM
        """
        if store_name in self.store and refresh == False:
            data = self.store[store_name]
            if model != None:
                return model(**data)
            return data

        if prefix == None and self.prefix == None:
            raise Exception("Prefix or parameter name didnt provided.")
        real_path = prefix or ""
        if self.prefix != None:
            real_path = f"{self.prefix}{real_path}"
        parameter = self.ssm.get_parameter(Name=real_path, WithDecryption=decryption)
        if store and store_name:
            self.store[store_name] = json.loads(parameter["Parameter"]["Value"])
        if model == None:
            return json.loads(parameter["Parameter"]["Value"])
        return model(**json.loads(parameter["Parameter"]["Value"]))



from marshmallow.fields import String
from marshmallow_objects.models import NestedModel
from zpy.api.models import ZModel

# Models 🚟


class DBSSMCredentialsModel(ZModel):
    uri = String(data_key="uri")
    port = String(data_key="port")
    service = String(data_key="service")
    user = String(data_key="user")
    password = String(data_key="pass")


class AESSSMModel(ZModel):
    aes_sk = String(data_key="aes_sk")


class AWSFirehoseModel(ZModel):
    access_key = String(data_key="access_key")
    secret_key = String(data_key="secret_key")


class AWSModel(ZModel):
    firehose = NestedModel(nested=AWSFirehoseModel)


class SSMConstants(ZModel):
    smm_prefix = String()
    local_db_ssm = String()
    db_smm = String()
    aes_sk = String()
    aws_ssm = String()



import json
from botocore.exceptions import ClientError
import boto3
from typing import Any, Dict
from . import AWSCredentials, CredentialsMode, AWS_DEFAULT_REGION


class S3:
    initialized: bool = False
    s3_client = None
    with_profile: bool = True

    def __init__(
        self,
        credentials: AWSCredentials = None,
        bucket: str = None,
        initialize: bool = False,
        with_profile: bool = True,
    ) -> None:
        self.credentials = credentials
        self.bucket = bucket
        self.with_profile = with_profile
        if initialize:
            self.__init_client(credentials, with_profile)

    def set_credentials_mode(self, mode: CredentialsMode):
        if mode == CredentialsMode.CREDENTIALS:
            self.with_profile = False
            return True
        self.with_profile = True
        return True

    def __init_client(self, credentials: AWSCredentials, profile: bool = True):
        if credentials is None and profile == False:
            raise Exception("Credentials didn't provided")
        if credentials != None and profile == False:
            self.s3_client = boto3.client(
                "s3",
                aws_access_key_id=credentials.access_key,
                aws_secret_access_key=credentials.secret_key,
                # aws_session_token=credentials.session_token,
                region_name=AWS_DEFAULT_REGION,
            )
        else:
            self.s3_client = boto3.client(
                "s3",
                region_name=AWS_DEFAULT_REGION,
            )
        self.initialized = True

    def get(
        self, full_key: str, bucket: str = None, credentials: AWSCredentials = None
    ) -> Any:
        """
        Get object from s3 bucket
        """
        real_bucket = self.__validate(bucket, credentials)
        s3_object = self.s3_client.get_object(Bucket=real_bucket, Key=full_key)
        content = s3_object["Body"].read()
        return content

    def download(
        self,
        full_key: str,
        bucket: str,
        local_file: str,
        credentials: AWSCredentials = None,
    ) -> Any:
        """
        Get object from s3 bucket
        """
        real_bucket = self.__validate(bucket, credentials)
        self.s3_client.download_file(real_bucket, full_key, local_file)
        return True

    def get_json(
        self, full_key: str, bucket: str = None, credentials: AWSCredentials = None
    ):
        json_obj = self.get(full_key, bucket, credentials)
        return json.loads(json_obj.decode("utf-8"))

    def put(
        self,
        object: Any,
        full_key: str,
        bucket: str = None,
        credentials: AWSCredentials = None,
    ) -> Any:
        """
        Put object from s3 bucket
        """
        real_bucket = self.__validate(bucket, credentials)
        result = self.s3_client.put_object(Body=object, Bucket=bucket, Key=full_key)
        return result

    def put_json(
        self,
        json_object: Dict,
        full_key: str,
        bucket: str = None,
        credentials: AWSCredentials = None,
    ):
        """
        Upload JSON|DICT object to S3

        Parameters
        ----------
        json_object : dict
            The dict json to upload s3
        full_key : str
            The file name
        bucket : str, optional
            Bucket name
        """
        json_parsed = str(json.dumps(json_object))
        return self.put(json_parsed, full_key, bucket, credentials)

    def upload(
        self,
        object_up: Any,
        object_name: str,
        bucket: str = None,
        credentials: AWSCredentials = None,
    ) -> bool:
        try:
            real_bucket = self.__validate(bucket, credentials)
            response = self.s3_client.upload_fileobj(
                object_up, real_bucket, object_name
            )
            return response
        except ClientError as e:
            print(e)
            return None

    def __validate(self, bucket: str = None, credentials: AWSCredentials = None) -> str:
        """
        Vrify aws credentials
        """
        real_bucket = self.bucket if bucket is None else bucket
        if real_bucket is None:
            raise Exception("Bucket didn't provided")
        if self.initialized == False:
            self.__init_client(self.credentials if credentials is None else credentials)
        return real_bucket




from enum import Enum


__author__ = "Noé Cruz | contactozurckz@gmail.com"
__copyright__ = "Copyright 2021, Small APi Project"
__credits__ = ["Noé Cruz", "Zurck'z"]
__license__ = "MIT"
__version__ = "0.0.1"
__maintainer__ = "Noé Cruz"
__email__ = "contactozurckz@gmail.com"
__status__ = "Dev"

AWS_DEFAULT_REGION = "us-east-1"
AWS_DNS = "https://s3.amazonaws.com/"


class AWSCredentials:
    def __init__(self, ak: str, sk: str, st: str = "", region: str = AWS_DEFAULT_REGION) -> None:
        self.access_key = ak
        self.secret_key = sk
        self.session_token = st
        self.region = region


class CredentialsMode(Enum):
    PROFILE = 1
    CREDENTIALS = 2


from marshmallow_objects import models
from datetime import datetime
from typing import Any, Callable, Dict, List, Optional
from copy import copy
import json


__author__ = "Noé Cruz | contactozurckz@gmail.com"
__copyright__ = "Copyright 2021, Small APi Project"
__credits__ = ["Noé Cruz", "Zurck'z"]
__license__ = "MIT"
__version__ = "0.0.1"
__maintainer__ = "Noé Cruz"
__email__ = "contactozurckz@gmail.com"
__status__ = "Dev"


#
# Zurck'z implementation
class ZModel(models.Model):
    """
    Zurckz Model
    """

    __remove_keys: List[str] = [
        "__dump_lock__",
        "__schema__",
        "__missing_fields__",
        "__setattr_func__",
        "_ZModel__remove_keys",
        "_ZModel__update_items",
    ]

    __update_items = {}

    def __init__(
        self,
        exclude: Optional[List[str]] = None,
        include: Optional[Dict[Any, Any]] = None,
        context=None,
        partial=None,
        **kwargs
    ):
        super().__init__(context=context, partial=partial, **kwargs)
        if exclude != None:
            self.__remove_keys = self.__remove_keys + exclude
        if include != None:
            self.__update_items = include

    def __str__(self):
        """
        Dump nested models by own properties
        """
        data = copy(self.__dict__)
        if self.__update_items != None:
            data.update(self.__update_items)
        [data.pop(k, None) for k in self.__remove_keys]
        for k in data.keys():
            if isinstance(data[k], models.Model):
                data[k] = json.loads(str(data[k]))
            elif isinstance(data[k], list):
                data[k] = [json.loads(str(it)) for it in data[k]]
            elif isinstance(data[k], datetime):
                data[k] = str(data[k])
        return json.dumps(data)

    def zdump(self):
        """
        Dump nested models by own properties
        """
        data = copy(self.__dict__)
        if self.__update_items != None:
            data.update(self.__update_items)
        [data.pop(k, None) for k in self.__remove_keys]
        for k in data.keys():
            if isinstance(data[k], models.Model):
                data[k] = json.loads(str(data[k]))
            elif isinstance(data[k], list):
                data[k] = [json.loads(str(it)) for it in data[k]]
            elif isinstance(data[k], datetime):
                data[k] = str(data[k])
        return data

    def sdump(
        self,
        exclude_keys: Optional[List[str]] = None,
        include: Optional[Dict[Any, Any]] = None,
        map: Optional[Callable[[Dict], Dict]] = None,
        map_args: Optional[List[Any]] = [],
        store_ex: bool = False,
        store_in: bool = False,
    ):
        """
        Model dump to json safely, checking the exclude key list

        Use this function instead of zdump.

        Parameters:
        -----------

        exclude_keys: List[str], Optional,
            List of string keys of exlude in dump process
        include: Dict[Any,Any], Optional,
            Object to include in model object after exclude process before of dump process
        map: Callable, Optional
            Callable function to tranform object after exclude and include process
        map_args: List[Any], Optional
            Argument list to passed to map callable function
        store_ex: bool, optional
            Indicate that the exclude key added to global model exclude key array
        store_in: bool, optional
            Indicate that the include object added to global model object
        """
        data = copy(self.__dict__)
        temp_exclude = copy(self.__remove_keys)
        if exclude_keys != None:
            temp_exclude = self.__remove_keys + exclude_keys
            if store_ex:
                self.__remove_keys = self.__remove_keys + exclude_keys
        [data.pop(k, None) for k in temp_exclude]
        temp_include = copy(self.__update_items)
        if include != None:
            temp_include.update(include)
            data.update(temp_include)
            if store_in:
                self.__update_items.update(include)
        else:
            if temp_include != None:
                data.update(temp_include)
        if map != None:
            data = map(data, *map_args)

        for k in data.keys():
            if isinstance(data[k], models.Model):
                data[k] = json.loads(str(data[k]))
            elif isinstance(data[k], list):
                # data[k] = [json.loads(str(it)) for it in data[k]]
                inner_list = []
                for it in data[k]:
                    if isinstance(it, str):
                        inner_list.append(it)
                    else:
                        inner_list.append(json.loads(str(it)))
                data[k] = inner_list
            elif isinstance(data[k], datetime):
                data[k] = str(data[k])
        return data

    def build(self):
        data = copy(self.__dict__)
        if self.__update_items != None:
            data.update(self.__update_items)
        [data.pop(k, None) for k in self.__remove_keys]
        return data



from zpy.api.stages import Stage
from io import StringIO
from enum import Enum
import logging

FORMAT = "ZR: %(asctime)s - %(name)s - %(levelname)s - %(message)s"
# logging.basicConfig(format=FORMAT)


class LColors(Enum):
    CEND = "\033[0m"
    CBLACK = "\33[30m"
    CRED = "\33[31m"
    CGREEN = "\33[32m"
    CYELLOW = "\33[33m"
    CBLUE = "\33[34m"
    CVIOLET = "\33[35m"
    CBEIGE = "\33[36m"
    CWHITE = "\33[37m"


def line(length: int = 100, char: str = "=") -> str:
    """
    Helper function
    """
    return "".join([char for x in range(length)])


def g_log(msg, phase: Stage):
    print(line())
    print(phase.title())
    print(phase.desc())
    print(msg)
    print(line())


def c_warn(msg, single=False, prefix: str = ""):
    if single:
        print(f"{LColors.CYELLOW.value}{prefix}{msg}{LColors.CEND.value}")
    else:
        print(f"{LColors.CYELLOW.value} [ZPy WARNING]: {msg}{LColors.CEND.value}")


def c_err(msg, single=False, prefix: str = ""):
    if single:
        print(f"{LColors.CRED.value}{prefix}{msg}{LColors.CEND.value}")
    else:
        print(f"{LColors.CRED.value} [ZPy ERROR]: {msg}{LColors.CEND.value}")


def c_info(msg, single=False, prefix: str = ""):
    if single:
        print(f"{LColors.CBLUE.value}{prefix}{msg}{LColors.CEND.value}")
    else:
        print(f"{LColors.CBLUE.value} [ZPy INFORMATION]: {msg}{LColors.CEND.value}")


def c_text(msg, single=False, prefix: str = ""):
    if single:
        print(f"{LColors.CWHITE.value}{prefix}{msg}{LColors.CEND.value}")
    else:
        print(f"{LColors.CWHITE.value} [ZPy]: {msg}{LColors.CEND.value}")


class TLogger:
    logger_name = "TradeLogger"

    def __init__(self, name=None, operation_id=None) -> None:
        self.log = logging.getLogger(self.logger_name if name is None else name)
        if self.log.handlers:
            for handler in self.log.handlers:
                self.log.removeHandler(handler)
        self.stream = StringIO()
        self.handler = logging.StreamHandler(self.stream)
        self.handler.setFormatter(logging.Formatter(FORMAT))
        self.log.addHandler(self.handler)
        self.log.setLevel(logging.DEBUG)
        self.operation_id = operation_id
        self.log.info(line())
        self.log.info("[ START OPERATION ID :: {0}]".format(self.operation_id))
        self.log.info(line())

    def add_error(self, error) -> None:
        self.log.error(error)

    def add_exception(self, exception) -> None:
        self.log.error(exception, exc_info=True)

    def add_warning(self, warning) -> None:
        self.log.error(warning)

    def add_info(self, info) -> None:
        self.log.info(info)

    def add_phase(self, phase: Stage):
        self.log.info(line())
        self.log.info("[ STARTING STAGE]")
        self.log.info(phase.title())
        self.log.info(phase.desc())
        self.log.info(line())

    def show_stack_trace(self):
        self.log.info(line())
        print(self.stream.getvalue())

    def dispose(self):
        self.handler.close()
        self.log.removeHandler(self.handler)


def add_log_to(log: any, logger: TLogger):
    if logger != None:
        if isinstance(log, Exception):
            logger.add_exception(log)
        else:
            logger.add_info(log)




from abc import abstractmethod


__author__ = "Noé Cruz | contactozurckz@gmail.com"
__copyright__ = "Copyright 2021, Small APi Project"
__credits__ = ["Noé Cruz", "Zurck'z"]
__license__ = "MIT"
__version__ = "0.0.1"
__maintainer__ = "Noé Cruz"
__email__ = "contactozurckz@gmail.com"
__status__ = "Dev"


class Validation():
    field: str = None
    rule: str = None
    message: str = None

    def __init__(self,field:str ,rule: str,error_messgae: str) -> None:
        self.field = field
        self.rule = rule
        self.message = error_messgae

    def json(self) -> dict:
        try:
            return vars(self)
        except:
            return self.__dict__

    @abstractmethod    
    def validate(self) -> bool:
        return False






if __name__ == '__main__':
    validations = [
        Validation(
            'name',
            "required",
            "El nombre es requerido"
        ),
        Validation(
            'address',
            "required",
            "El nombre es requerido"
        ),
        Validation(
            'age',
            "required|number|max:40,min:21",
            "El nombre es requerido"
        )
    ]
    
    v = Validation(
            'age',
            "required|number|max:40,min:21",
            "El nombre es requerido"
        ).json()
    print(v)
    


from abc import ABC, abstractmethod
from zpy.logger import TLogger
from typing import Any, Dict, Optional, TypeVar
from marshmallow_objects import models


T = TypeVar("T")


class DBConnection(ABC):
    @abstractmethod
    def init_local_client(self, path: str):
        """
        Initialize local client
        """
        pass

    @abstractmethod
    def connect(self):
        pass

    @abstractmethod
    def close(self):
        pass

    @abstractmethod
    def is_connected(self):
        pass

    @abstractmethod
    def get_connection(self):
        pass

    @abstractmethod
    def execute(
        self,
        function: str,
        type,
        parameters: dict = None,
        pool_connection=None,
    ):
        pass

    @abstractmethod
    def get_pool_connection(self, logger: TLogger):
        pass

    @abstractmethod
    def release_connection(self, connection) -> bool:
        pass

    @abstractmethod
    def call(self, fn: str, type, params, schmea: T):
        pass

    @abstractmethod
    def close_pool(self):
        pass

    @abstractmethod
    def initialize_pool(
        self,
        dns=None,
        homogeneous: bool = False,
        max: int = 5,
        user=None,
        pwd=None,
        min: int = 1,
        threaded: bool = False,
    ):
        pass

    @abstractmethod
    def exec(
        self,
        fn: str,
        ret_type: Any,
        params: Optional[Dict] = None,
        model: Optional[models.Model] = None,
        connection=None,
    ):
        """
        New feature for call oracle db functions
        Use this function instead function 'call'

        Parameters
        ----------
        fn : str | required
            Function name with package name: PO_LINES_PKG.FN_GET_LINE

        ret_type : OracleType | required
            The return type of oracle db function

        params : Dict | Optional
            Set parameter that the oracle funtion expects

        model : marshmallow_objects.models.Model | Optional
            Model specification where the db data will be volcated

        connection : DB Connection | Optional
            The db connection object, if it is not passed by params, it tries to get a global instance

        Raises
        ------
        NotValueProvided
            Connection

        Returns
        -------
        result set : Union[List[Model],int,float,str]
            The result set of oracle db function
        """
        pass


from typing import Any, Dict, List
from zpy.cloud.aws import AWSCredentials, AWS_DEFAULT_REGION
import boto3
import json


class Firehose:
    fh_client = None
    with_profile: bool = False

    def __init__(
        self,
        credentials: AWSCredentials = None,
        with_profile: bool = True,
    ) -> None:
        self.with_profile = with_profile
        if with_profile or credentials is None:
            self.fh_client = boto3.client("firehose", region_name=AWS_DEFAULT_REGION)
        else:
            self.fh_client = boto3.client(
                "firehose",
                aws_access_key_id=credentials.access_key,
                aws_secret_access_key=credentials.secret_key,
                region_name=credentials.region,
            )

    def send_data_record(self, data: dict, stream_name: str):
        """
        Put one record to delivery stream
        """
        try:
            response = self.fh_client.put_record(
                DeliveryStreamName=stream_name, Record=self.__prepare_record(data)
            )
            return response
        except Exception as e:
            print(e)
            return None

    def __prepare_record(self, data: Dict) -> Dict:
        dumped = json.dumps(data)
        encoded = dumped.encode("utf-8")
        return {"Data": encoded}

    def send_batch_data(self, data: List[Dict[Any, Any]], stream_name: str):
        """
        Put one record to delivery stream
        :data List of record to send firehouse
        """
        try:
            records = list(map(self.__prepare_record, data))
            response = self.fh_client.put_record_batch(
                DeliveryStreamName=stream_name, Records=records
            )
            return response
        except Exception as e:
            print(e)
            return None

    def describe_stream(
        self, stream_name: str, limit: int = 123, startId: str = None
    ) -> dict:
        try:
            response = self.fh_client.describe_delivery_stream(
                DeliveryStreamName=stream_name,
                Limit=limit,
                ExclusiveStartDestinationId=startId,
            )
            return response
        except Exception as e:
            print(e)
            return None


from typing import Union
import boto3


class SQS:
    queue_name = None
    sqs_client = None
    queue_object = None
    gloabl_initialized: bool = False

    def __init__(self, queue_url: str = None) -> None:
        self.sqs_client = boto3.client("sqs")
        if queue_url != None:
            self.queue_name = queue_url
            # self.__sqs_initializer(name=queue_url, is_global=True)

    def __sqs_initializer(self, name: str = None, is_global: bool = False):
        sqs_name = self.__validate(name=name)
        sqs_object = self.sqs_client.get_queue_by_name(QueueName=sqs_name)
        if is_global:
            self.queue_name = sqs_name
            self.queue_object = sqs_object
            self.gloabl_initialized = True
            return self.queue_object
        return sqs_object

    def __validate(self, name: str = None) -> Union[str, None]:
        """
        Validate sqs name and return sqs object acording name
        """
        real_name = self.queue_name if name == None else name
        if real_name == None:
            raise Exception("SQS Name didn't provide")
        return real_name

    def send(self, msg: str, queue_url: str = None, attributes: dict = {}):
        real_url = self.queue_name if queue_url == None else queue_url
        if real_url == None:
            raise Exception("SQS URL didn't provided")
        response = self.sqs_client.send_message(
            QueueUrl=queue_url,
            DelaySeconds=0,
            MessageAttributes=attributes,
            MessageBody=msg,
        )
        return response

    def __send_message(self, msg: str, queue_url: str = None):
        """
        Send MESSAGE to SQS
        IN DEVELOP STAGE
        """
        client = self.__sqs_initializer(queue_url)



from zpy.logger import TLogger, c_info
from zpy.db.utils import get_current_schema
from marshmallow_objects import models
from typing import Any, Dict, List, Optional, Union
from zpy.db import DBConnection
from zpy.utils.funcs import exec_ifnt_null, safely_exec
from enum import Enum
from marshmallow import Schema
import cx_Oracle
import json
import logging

# from . import T


__author__ = "Noé Cruz | contactozurckz@gmail.com"
__copyright__ = "Copyright 2021, Small APi Project"
__credits__ = ["Noé Cruz", "Zurck'z"]
__license__ = "MIT"
__version__ = "0.0.1"
__maintainer__ = "Noé Cruz"
__email__ = "contactozurckz@gmail.com"
__status__ = "Dev"


class OracleType(Enum):
    cursor = cx_Oracle.CURSOR
    number = cx_Oracle.NUMBER
    string = cx_Oracle.STRING
    integer = cx_Oracle.NUMBER
    decimal = cx_Oracle.NUMBER


class OracleParam(Enum):
    LIST_INTEGR = "LIST_INTEGER"
    LIST_STR = "LIST_VARCHAR"
    LIST_CLOB = "LIST_CLOB"


class ZParam:
    def __init__(
        self,
        value: Union[List[int], List[float], List[str], List[Any]],
        paramType: OracleParam,
        key: str,
        origin: str = None,
    ) -> None:
        self.value = value
        self.paramType = paramType
        self.key = key
        self.origin = origin


class IntList(ZParam):
    def __init__(self, value: List[int], key: str, origin: str = None) -> None:
        super().__init__(value, OracleParam.LIST_INTEGR, key, origin)


class StrList(ZParam):
    def __init__(self, value: List[str], key: str, origin: str = None) -> None:
        super().__init__(value, OracleParam.LIST_STR, key, origin)


class ClobList(ZParam):
    def __init__(self, value: List[Any], key: str, origin: str) -> None:
        super().__init__(value, OracleParam.LIST_CLOB, key, origin)


class OracleDB(DBConnection):
    __local_client_initialized: bool = False
    __local_client_path: str = None
    __config_connection: dict = None
    __connection = None
    __is_connected: bool = False
    __pool = None
    __pool_created: bool = False
    __schemas: List[Dict] = None
    __env: str = None
    __verbose: bool = False

    # * Pool configurations
    __max: int = 5
    __min: int = 1
    __threading: bool = False
    __homogeneus: bool = True

    def __init__(
        self,
        config: dict = None,
        local_client_path: str = None,
        schemas: List[Dict] = None,
        env: str = None,
        verbose: bool = False,
    ) -> None:
        self.__local_client_path = local_client_path
        self.__config_connection = config
        self.__schemas = schemas
        self.__env = env
        self.__verbose = verbose

    def init_local_client(self, path: str):
        if self.__local_client_initialized:
            return
        value = path if self.__local_client_path is None else self.__local_client_path
        try:
            if value is None:
                raise Exception("Local client path not provided.")
            cx_Oracle.init_oracle_client(lib_dir=value)
            self.__local_client_initialized = True
        except Exception as e:
            self.__local_client_initialized = False
            logging.exception(e)

    def __data_connection_checker(self, config: dict = None, mode="TSN") -> str:
        values = (
            config if self.__config_connection is None else self.__config_connection
        )
        if values is None:
            raise Exception("The data for the connection was not provided")
        server = values["DB_URI"]
        port = values["DB_PORT"]
        service = values["DB_SERVICE"]
        user = values["DB_USER"]
        password = values["DB_PASSWORD"]
        if mode == "DSN":
            return "{0}:{1}/{2}".format(server, port, service)
        return "{0}/{1}@{2}:{3}/{4}".format(user, password, server, port, service)

    def get_tsn_dsn_conenction(self, config: dict, mode="TSN") -> str:
        return self.__data_connection_checker(config, mode)

    def connect(self, config: dict = None):
        """
        Start oracle connection
        """
        if self.__is_connected:
            return True
        try:
            str_connection = self.__data_connection_checker(config)
            self.__connection = cx_Oracle.connect(str_connection)
            self.__is_connected = True
            return True
        except Exception as e:
            raise e

    def close(self):
        if self.__is_connected:
            self.__connection.close()
            self.__is_connected = False
            self.is_connected = False

    def is_connected(self) -> bool:
        return self.__is_connected

    def get_connection(self):
        return self.__connection

    def initialize_pool(
        self,
        dns=None,
        homogeneous: bool = False,
        max: int = 5,
        user=None,
        pwd=None,
        min: int = 1,
        threaded: bool = False,
    ) -> bool:
        if self.__pool_created or dns is None:
            return False
        print("Initializing Pool")
        self.__pool = cx_Oracle.SessionPool(
            user=user,
            password=pwd,
            dsn=dns,
            homogeneous=homogeneous,
            encoding="UTF-8",
            max=max,
            min=min,
            threaded=threaded,
        )
        self.__pool_created = True
        self.__min = min
        self.__max = max
        self.__threading = threaded
        self.__homogeneus = homogeneous
        print("Pool Started Successfuly")
        return True

    def close_pool(self, force: bool = False):
        try:
            if self.__pool_created:
                self.__pool.close(force=force)
                self.__pool_created = False
                print("Pool Closed Successfuly")
        except Exception as e:
            logging.exception(e)

    def get_pool_connection(self, logger: TLogger = None):
        if self.__pool_created:
            return self.__pool.acquire()
        exec_ifnt_null(
            lambda l: l.add_info("DB POOL NOT INITILIZED, TRY INITIALIZE AGAIN"),
            args=[logger],
        )
        try:
            self.initialize_pool(
                dns=self.get_tsn_dsn_conenction(self.__config_connection, "DSN"),
                homogeneous=self.__homogeneus,
                user=self.__config_connection["DB_USER"],
                pwd=self.__config_connection["DB_PASSWORD"],
                max=self.__max,
                min=self.__min,
                threaded=self.__threading,
            )
            self.__pool_created = True
            return self.__pool.acquire()
        except Exception as e:
            logger.add_exception(e)
            self.__pool_created = False

    def release_connection(self, connection) -> bool:
        try:
            if self.__pool_created:
                self.__pool.release(connection)
                return True
        except Exception as e:
            print("CATCH EXCEPTION WHEN TRY RELEASE POOL CONNECTION")
            logging.exception(e)
            return False

    def __proccess_result(self, result_set, type: OracleType, schema: Schema = None):
        if type == OracleType.cursor:
            columns = [field[0] for field in result_set.description]
            if schema is None:
                columns = [field[0] for field in result_set.description]
                rows = [d for d in result_set]
                data = [dict(zip(columns, row)) for row in rows]
                for d in data:
                    for key, value in d.items():
                        if isinstance(d[key], cx_Oracle.LOB):
                            d[key] = json.loads(str(value))
                return {"hasError": False, "data": json.dumps(data, default=str)}
            else:
                # [print(dict(zip(columns, r))) for r in result_set]
                return [schema.load(dict(zip(columns, r))) for r in result_set]
        elif OracleType.number:
            try:
                return float(result_set)
            except:
                return result_set
        else:
            return str(result_set)

    def execute(
        self,
        function: str,
        type: OracleType,
        parameters: dict = None,
        pool_connection=None,
    ):
        """
        Execute or call oracle functions - FN v0.0.1 | Core v0.0.1
        """
        if pool_connection is not None:
            cursor = pool_connection.cursor()
        else:
            cursor = self.__connection.cursor()
        if self.__verbose:
            self.show_info(function, parameters, type, None, None, None)
        try:
            db_execute = (
                cursor.callfunc(function, type.value, keywordParameters=parameters)
                if parameters != None
                else cursor.callfunc(function, type.value)
            )
            if type == OracleType.cursor:
                columns = [field[0] for field in db_execute.description]
                rows = [d for d in db_execute]
                data = [dict(zip(columns, row)) for row in rows]
                for d in data:
                    for key, value in d.items():
                        if isinstance(d[key], cx_Oracle.LOB):
                            d[key] = json.loads(str(value))
                db_dto = {"hasError": False, "data": json.dumps(data, default=str)}
            elif OracleType.number:
                db_dto = {"hasError": False, "data": str(db_execute)}
            else:
                db_dto = {"hasError": False, "data": db_execute}
        except Exception as e:
            db_dto = {"hasError": True, "data": str(e)}
        safely_exec(lambda c: c.close(), args=[cursor])  # * Close cursor
        return db_dto

    def call(
        self,
        fn: str,
        type: OracleType,
        params: dict,
        schema: Schema = None,
        pool_connection=None,
    ):
        """
        Execute or call oracle functions - FN v0.0.1 | Core v0.0.2
        """
        if pool_connection is not None:
            cursor = pool_connection.cursor()
        else:
            cursor = self.__connection.cursor()

        if self.__verbose:
            self.show_info(fn, params, type, schema, None, None)

        result_set = (
            cursor.callfunc(fn, type.value, keywordParameters=params)
            if params != None
            else cursor.callfunc(fn, type.value)
        )
        safely_exec(lambda c: c.close(), args=[cursor])  # * Close cursor
        return self.__proccess_result(result_set, type, schema)

    def exec(
        self,
        fn: str,
        ret_type: OracleType,
        params: Optional[Dict] = None,
        custom_params: Optional[List[ZParam]] = None,
        model: Optional[models.Model] = None,
        connection=None,
        db_schema: str = None,
        env: str = None,
    ):
        """
        Execute or call oracle functions - FN v0.0.1 | Core v0.0.7

        New feature for call oracle db functions
        Use this function instead function 'call'

        Parameters
        ----------
        fn : str | required
            Function name with package name: PO_LINES_PKG.FN_GET_LINE

        ret_type : OracleType | required
            The return type of oracle db function

        params : Dict | Optional
            Set parameter that the oracle funtion expects

        custom_params : Optional[List[ZParam, IntList, StrList, ClobList]] | Optional
            Custom Set parameter that the oracle funtion expects, see avaialble custom types

        model : marshmallow_objects.models.Model | Optional
            Model specification where the db data will be volcated

        connection : DB Connection | Optional
            The db connection object, if it is not passed by params, it tries to get a global instance

        Raises
        ------
        NotValueProvided
            Connection

        Returns
        -------
        result set : Union[List[Model],int,float,str]
            The result set of oracle db function
        """
        cursor = None
        if connection is not None:
            cursor = connection.cursor()
        else:
            cursor = self.__connection.cursor()
        if connection is None:
            raise Exception("Can't get db connection")

        if db_schema is None and self.__schemas is not None:
            db_schema = get_current_schema(self.__schemas, env, self.__env)

        if custom_params != None and len(custom_params) > 0:
            if params == None:
                params = {}
            # * Find the current env for extract the schema
            for custom in custom_params:
                params[custom.key] = self.__custom_param(
                    connection,
                    paramType=custom.paramType,
                    value=custom.value,
                    schema=db_schema,
                )

        fn = (
            fn
            if db_schema is None or db_schema.replace(" ", "") == ""
            else f"{db_schema}.{fn}"
        )
        if self.__verbose:
            self.show_info(fn, params, ret_type, model, db_schema, env)
        result_set = (
            cursor.callfunc(fn, ret_type.value, keywordParameters=params)
            if params != None
            else cursor.callfunc(fn, ret_type.value)
        )
        safely_exec(lambda c: c.close(), args=[cursor])
        return self.__proccess_result_set(result_set, ret_type, model)

    def __proccess_result_set(
        self, result_set, ret_type: OracleType, model: models.Model = None
    ):
        """
        New version of result set processor
        """
        if ret_type == OracleType.cursor:
            columns = [field[0] for field in result_set.description]
            if model is None:
                columns = [field[0] for field in result_set.description]
                rows = [d for d in result_set]
                data = [dict(zip(columns, row)) for row in rows]
                for d in data:
                    for key, value in d.items():
                        if isinstance(d[key], cx_Oracle.LOB):
                            d[key] = json.loads(str(value))
                return {"data": json.dumps(data, default=str)}
            else:
                return [model(**dict(zip(columns, r))) for r in result_set]
        elif OracleType.number:
            try:
                return float(result_set)
            except:
                return result_set
        elif OracleType.integer:
            try:
                return int(result_set)
            except:
                return result_set
        elif OracleType.decimal:
            try:
                return float(result_set)
            except:
                return result_set
        else:
            return str(result_set)

    def __custom_param(
        self,
        connection: Any,
        paramType: OracleParam,
        value: Union[List[int], List[float], List[str], List[Any]],
        schema: str = None,
    ):
        """
        Make custom param
        """
        db_schema = (
            "" if (schema is None or schema.replace(" ", "") == "") else f"{schema}."
        )
        list_type = connection.gettype(f"{db_schema}{paramType.value}")
        return list_type.newobject(value)

    def show_info(self, fn, params, ret_type, v_model, curr_schema, l_env):
        c_info("\n|-------------------------------------------------|\n", True)
        c_info(f"    Function Called: {fn} ", True)
        c_info("             Params: {}".format(params), True)
        c_info("        Return Type: {}".format(ret_type.value), True)
        c_info(f" Ref Volcated Model: {v_model}", True)
        c_info(f"          DB Schema: {curr_schema}", True)
        c_info(f"        Environment: P: {l_env} G: {self.__env}", True)
        c_info("\n|-------------------------------------------------|\n", True)



from enum import Enum
from typing import Dict
from zpy.api.stages import Unspecified


__author__ = "Noé Cruz | contactozurckz@gmail.com"
__copyright__ = "Copyright 2021, Small APi Project"
__credits__ = ["Noé Cruz", "Zurck'z"]
__license__ = "MIT"
__version__ = "0.0.1"
__maintainer__ = "Noé Cruz"
__email__ = "contactozurckz@gmail.com"
__status__ = "Dev"


class ErrorDomain(Enum):
    GLOBAL = ("GLOBAL", "")
    IN_PHASE = ("IN PHASE", "")
    INVALID_REQUEST = ("INVALID REQUEST", "")
    INVALID_PARAMETER = ("INVALID PARAMETER", "")


class ErrorCodes(Enum):
    UNSPECIFIED = 99
    GENERAL = 100
    BAD_REQUEST = 101
    INVALID_TYPE_REQUEST = 102
    DB_OPERATION = 103
    SERVICE_PROCESS = 104


# DOMAIN_ERROR_KEY   = 'domain'
REASON_ERROR_KEY = "reason"
MESSAGE_ERROR_KEY = "message"
METADATA_KEY = "metadata"
CODE_ERROR_KEY = "code"
# STAGE_ERROR_KEY    = 'stage'


class ErrorBuilder(object):
    error = {
        # DOMAIN_ERROR_KEY  : ErrorDomain.GLOBAL,
        REASON_ERROR_KEY: "",
        MESSAGE_ERROR_KEY: "",
        CODE_ERROR_KEY: "",
        # STAGE_ERROR_KEY    : Unspecified().name,
        METADATA_KEY: None,
    }

    @staticmethod
    def common(reason: str, message: str, code: ErrorCodes = ErrorCodes.GENERAL, meta=None):
        return {
            # DOMAIN_ERROR_KEY: domain.value[0],
            REASON_ERROR_KEY: reason,
            MESSAGE_ERROR_KEY: message,
            CODE_ERROR_KEY: code.value,
            METADATA_KEY: meta,
            # STAGE_ERROR_KEY: stage.name,
        }

    # def add_stage(self, stage:Stage = Unspecified()):
    #     self.error[STAGE_ERROR_KEY] = stage.name
    #     return self

    def add_meta(self, meta: Dict = Unspecified()):
        self.error[METADATA_KEY] = meta
        return self

    # def add_domain(self, domain:ErrorDomain = ErrorDomain.GLOBAL):
    #     self.error[DOMAIN_ERROR_KEY] = domain.value[0]
    #     return self

    def add_reason(self, reason: str = ""):
        self.error[REASON_ERROR_KEY] = reason
        return self

    def add_message(self, message: str = ""):
        self.error[MESSAGE_ERROR_KEY] = message
        return self

    def add_code(self, code: ErrorCodes = ErrorCodes.UNSPECIFIED):
        self.error[CODE_ERROR_KEY] = code
        return self

    def build(self) -> dict:
        return self.error


from enum import Enum

class EnvironmentKeys(Enum):
    ENVIRONMENT = 'env'

    def __str__(self) -> str:
        return self.value
class Messages(Enum):
    GEN_MSG_ERROR = "[GE-00] An unspecified error was generated while the request was being processed."

    def __str__(self):
        return self.value


from enum import IntEnum
from zpy.api.hooks import NotFoundHook, ZHook
from zpy.api.middleware import ZMiddleware, ParserMiddleWare
from functools import wraps
from typing import Any, Dict, Generic, List, Optional, TypeVar
from flask.views import MethodView
from flask_restful import Api
from flask import Flask
from flask_cors import CORS


__author__ = "Noé Cruz | contactozurckz@gmail.com"
__copyright__ = "Copyright 2021, Small APi Project"
__credits__ = ["Noé Cruz", "Zurck'z"]
__license__ = "MIT"
__version__ = "0.0.1"
__maintainer__ = "Noé Cruz"
__email__ = "contactozurckz@gmail.com"
__status__ = "Dev"

# Custom Type for typing generic functions
T = TypeVar("T")

##  Class ApiConfiguration
#   Multiple environment variables for microservices
#
class ApiConfiguration:
    DEBUG: bool = False
    TESTING: bool = False
    CSRF_ENABLED: bool = False
    SECRET: str = ""
    DB_URI: str = ""
    DB_SERVICE: str = ""
    DB_PORT: str = ""
    DB_USER: str = ""
    DB_PASSWORD: str = ""
    ENVIRONMENT: str = ""
    CUSTOM: Dict = {}

    def to_dict(self) -> dict:
        return {
            "DEBUG": self.DEBUG,
            "TESTING": self.TESTING,
            "CSRF_ENABLED": self.CSRF_ENABLED,
            "SECRET": self.SECRET,
            "DB_URI": self.DB_URI,
            "DB_SERVICE": self.DB_SERVICE,
            "DB_PORT": self.DB_PORT,
            "DB_USER": self.DB_USER,
            "DB_PASSWORD": self.DB_PASSWORD,
            "ENVIRONMENT": self.ENVIRONMENT,
            "CUSTOM": self.CUSTOM,
        }


##  Class ZResource | Zurck'z Resources
#   Resource class, contains resources and path related to it
#
class ZResource:
    """
    Micro service resource for api.
    """

    def __init__(self, path: str, resource: MethodView, **kwargs) -> None:
        self.resource = resource
        self.path = path
        self.kwargs = kwargs


def create_app(
    config: ApiConfiguration,
    resources: List[ZResource],
    main_path: str,
    path_cors_allow=None,
) -> Flask:
    """
    API Builder
    """
    app = Flask(__name__, instance_relative_config=True)
    app.config.from_object(config)
    path_allow = path_cors_allow
    if path_cors_allow is None:
        path_allow = r"{}{}".format(main_path, "/*")
    CORS(app, resources={path_allow: {"origins": "*"}})
    api = Api(app)
    [
        api.add_resource(
            r.resource,
            "{0}{1}".format(main_path, r.path),
            resource_class_kwargs=r.kwargs,
        )
        for r in resources
    ]
    return app


def api(
    base: str = "",
    config: ApiConfiguration = None,
    middlewares: List[ZMiddleware] = [],
    middlewares_args: List[dict] = [],
    hooks: List[ZHook] = [],
    hooks_args: List[dict] = [],
    path_cors_allow=None,
) -> Flask:
    """
    Micro Service Setup
    """
    if config is None:
        raise Exception("Api Configurations were not provided")

    def callable(api_definitor) -> Flask:
        @wraps(api_definitor)
        def wrapper(*args, **kwargs) -> Flask:
            resources: List[ZResource] = api_definitor(*args, **kwargs)
            app: Flask = create_app(config, resources, base, path_cors_allow)
            middlewares.append(ParserMiddleWare)
            middlewares_args.append({})
            hooks.append(NotFoundHook)
            hooks_args.append({})
            for i, m in enumerate(middlewares):
                args = middlewares_args[i] if i < len(middlewares_args) else {}
                args.update(config.CUSTOM)
                app.wsgi_app = m(app.wsgi_app, **args)
            for i, h in enumerate(hooks):
                args = hooks_args[i] if i < len(hooks_args) else {}
                args.update(config.CUSTOM)
                h().execute(app, **args)
            return app

        return wrapper

    return callable


class ZContractStatus(IntEnum):
    ERROR = -1
    SUCCESS = 1
    PENDING = 0


class ZContract(Generic[T]):

    data: T = None
    status: ZContractStatus = None
    errors: List[Dict] = None

    def __init__(
        self,
        data: Optional[T] = None,
        status: ZContractStatus = ZContractStatus.PENDING,
        errors: Optional[List[Any]] = None,
    ) -> None:
        super().__init__()
        self.data = data
        self.status = status
        self.errors = errors



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="zpy-flask-msc",
    version="0.0.1",
    author="Noé Cruz | linkedin.com/in/zurckz/",
    author_email="zurckz.services@gmail.com",
    description="Helper layer for apis development with Python and Flask",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/NoeCruzMW",
    packages=setuptools.find_packages(),
    install_requires=[
        "pycryptodome",
        "pycryptodomex",
        "Flask",
        "flask-rest-api",
        "Flask-RESTful",
        "marshmallow",
        "python-dotenv",
        "flask_cors",
        "cx-Oracle",
        "marshmallow_objects"
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
)


from typing import Dict
from zpy.api import ApiConfiguration


class DevConfigurations(ApiConfiguration):
    DEBUG: bool = True
    TESTING: bool = True
    CSRF_ENABLED: bool = False
    SECRET: str = ""
    DB_URI: str = ""
    DB_SERVICE: str = ""
    DB_PORT: str = ""
    DB_USER: str = ""
    DB_PASSWORD: str = ""
    ENVIRONMENT: str = "local_dev"
    CUSTOM: Dict = {}


apiConfig = DevConfigurations()


from zpy.custom import Plugin
from zpy.api import ApiConfiguration
from zpy.cloud.aws.ssm import SSMParameter
from zpy.cloud.aws.firehose import Firehose
from zpy.db.oracle import OracleDB
import os


class PFirehose(Plugin):
    def initialize(self, config: ApiConfiguration, ssm: SSMParameter, **kwargs):
        return Firehose(with_profile=True)


class POracle(Plugin):
    """"""

    def initialize(self, config: ApiConfiguration, ssm: SSMParameter, **kwargs):

        local_client = None if "local_client" not in kwargs else kwargs["local_client"]
        verbose = False if "verbose" not in kwargs else kwargs["verbose"]
        schemas = None if "schemas" not in kwargs else kwargs["schemas"]
        pool_initialize = (
            False if "pool_initialize" not in kwargs else kwargs["pool_initialize"]
        )
        db = OracleDB(
            config=config.to_dict(),
            local_client_path=local_client,
            schemas=schemas,
            env=config.ENVIRONMENT,
            verbose=verbose,
        )
        if pool_initialize:
            try:
                if config.ENVIRONMENT == "localdev":
                    db.init_local_client(os.getenv("LOCAL_CLIENT"))
                db.initialize_pool(
                    dns=db.get_tsn_dsn_conenction(config.to_dict(), "DSN"),
                    homogeneous=True,
                    user=config.DB_USER,
                    pwd=config.DB_PASSWORD,
                )
            except Exception as e:
                print(e)
        return db


from abc import ABC, abstractmethod
from zpy.api.reponse import Builder
from zpy.api.stages import Decrypt
from zpy.logger import g_log
from zpy.api.errors import ErrorBuilder
from flask import Flask
from typing import Any
from flask.wrappers import Request, Response
from zpy.utils.Encryptor import AESEncryptor
import json


__author__ = "Noé Cruz | contactozurckz@gmail.com"
__copyright__ = "Copyright 2021, Small APi Project"
__credits__ = ["Noé Cruz", "Zurck'z"]
__license__ = "upax"
__version__ = "0.0.1"
__maintainer__ = "Noé Cruz"
__email__ = "contactozurckz@gmail.com"
__status__ = "Dev"


# Middlewares | Zurck'Z Middlware
# Base middleware for flask
class ZMiddleware(ABC):
    def __init__(self, app: Flask, **kwargs) -> None:
        super().__init__()
        self.app = app
        self.kwargs = kwargs

    @abstractmethod
    def __call__(self, environ: Any, start_response: Any) -> Any:
        return self.app(environ, start_response)


# Custom Middlewares
# Encrypt body of responses with AES algorithm
class EncryptMiddleWare(ZMiddleware):
    def __init__(self, app: Flask, **kwargs) -> None:
        super().__init__(app, **kwargs)
        self.app = app

    def __call__(self, environ: Any, start_response: Any) -> Any:
        response = Response(environ)
        return super().__call__(environ, start_response)
        # return response(environ,start_response)


# Custom Middlewares
# Encrypt body of responses with AES algorithm
class DecryptMiddleWare(ZMiddleware):
    def __init__(self, app: Flask, **kwargs):
        super().__init__(app, **kwargs)
        self.app = app

    def __call__(self, environ, start_response):
        try:
            if environ["request"]:
                aes: str = None
                if (
                    self.kwargs != None and "aes_sk" in self.kwargs
                ):  # ! WARNING HARD KEY FOR EXTARCT AES SK
                    aes = self.kwargs["aes_sk"]
                encrypt_data = environ["request"]
                decrypt_data = AESEncryptor.decrypt_ws_response(
                    encrypt_data, secret_key=aes
                )
                environ["request"] = decrypt_data
            return self.app(environ, start_response)
        except Exception as e:
            stage = Decrypt()
            g_log(e, stage)
            res = Response(
                json.dumps(
                    Builder.error(
                        errors=[
                            ErrorBuilder().common(
                                "Threw exception on decrypt process",
                                "Request supplied not have a valid format",
                                stage,
                            )
                        ]
                    )
                ),
                mimetype="text/json",
                status=500,
            )
            return res(environ, start_response)


class ParserMiddleWare(ZMiddleware):
    """
    Default middleware for custom access response
    """

    def __init__(self, app: Flask, **kwargs):
        super().__init__(app, **kwargs)
        self.app = app

    def __call__(self, environ, start_response):
        request = Request(environ)
        try:
            if request.data:
                environ["request"] = json.loads(request.data)
            else:
                environ["request"] = None
            return self.app(environ, start_response)
        except Exception as e:
            stage = Decrypt()
            g_log(e, stage)
            res = Response(
                json.dumps(
                    Builder.error(
                        errors=[
                            ErrorBuilder().common(
                                "Threw exception on decrypt process",
                                "Request supplied not have a valid format",
                                stage,
                            )
                        ]
                    )
                ),
                mimetype="text/json",
                status=500,
            )
            return res(environ, start_response)



from zpy.api.resource import ZResource, HTTP_METHODS


class GreetingResource(ZResource):

    blocked_methods = [
        HTTP_METHODS.POST,
        HTTP_METHODS.DELETE,
        HTTP_METHODS.PATCH,
        HTTP_METHODS.PUT,
    ]

    def __init__(self, **kwargs) -> None:
        super().__init__()

    def get(self):
        l, i = super().new_operation()
        try:
            return self.success({"greeting": "hello world!"}, logger=l)
        except Exception as e:
            return self.handle_exceptions(e, l, i)



from zpy.api import ZContract, ZContractStatus
from zpy.logger import TLogger, add_log_to, c_err, c_text
from zpy.api.exceptions import ZDBIntegrityError, ZDatabaseOperationError, ZError
from typing import Any, Dict, List, Optional, TypeVar
from zpy.db import DBConnection
from zpy.api.stages import DataAccess
from zpy.db.oracle import OracleType, ZParam
from zpy.db.transaction import ZOperation, ZTranstaction
import logging
import cx_Oracle

__author__ = "Noé Cruz | contactozurckz@gmail.com"
__copyright__ = "Copyright 2021, Small APi Project"
__credits__ = ["Noé Cruz", "Zurck'z"]
__license__ = "MIT"
__version__ = "0.0.1"
__maintainer__ = "Noé Cruz"
__email__ = "contactozurckz@gmail.com"
__status__ = "Dev"


T = TypeVar("T")


class RepositoryResult(ZContract[T]):
    def __init__(
        self,
        data: T = None,
        status: ZContractStatus = ZContractStatus.PENDING,
        errors: List[Any] = None,
    ) -> None:
        super().__init__(data=data, status=status, errors=errors)


class ZRepository(object):
    db: DBConnection
    logger: TLogger = None
    transact: ZTranstaction

    def __init__(self, db: DBConnection = None) -> None:
        super().__init__()
        self.db = db
        self.transact = None

    def _check_transact(self):
        if self.transact == None:
            raise "Transactionable component did not initialize."

    def set_transaction_dispatcher(self, transact: ZTranstaction):
        self.transact = transact

    def set_logger(self, logger: TLogger = None):
        self.logger = logger
        if self.logger != None:
            self.logger.add_phase(DataAccess())

    def release_db_connection(self, connection):
        if self.db != None:
            return self.db.release_connection(connection)
        raise ZDatabaseOperationError("DB Connection didn't provided")

    def get_db_connection(self, logger: TLogger = None):
        if self.db != None:
            return self.db.get_pool_connection(logger=logger)
        raise ZDatabaseOperationError("DB Connection didn't provided")

    def __show_db_error(
        self,
        e: Exception,
        fn: str,
        type_err: str,
        native_params: Dict = None,
        custom_params: Optional[List[ZParam]] = None,
        logger: TLogger = None,
    ):
        """
        Show error acording logger value, if logger is null print on console

        RETURNS:
        -------
            added into logger

        """
        logger = self.logger if logger == None else logger
        if logger == None:
            c_err("\n\tAN ERROR OCURRED WHEN EXECUTE DB FUNCTION\n", True)
            c_text(f"     Function: {fn}", True)
            c_text(f"         Type: {type_err}", True)
            c_text("  With params: {}".format(native_params), True)
            c_text("Custom params: {}".format(custom_params), True)
            c_text("--------------------------------------------------", True)
            c_text("Details:", True)
            logging.exception(e)
            c_err("--------------------------------------------------", True)
            return False
        else:
            add_log_to("\n\tAN ERROR OCURRED WHEN EXECUTE DB FUNCTION\n", logger)
            add_log_to(f"     Function: {fn}", logger)
            add_log_to(f"         Type: {type_err}", logger)
            add_log_to(f"  With params: {native_params}", logger)
            add_log_to(f"Custom params: {custom_params}", logger)
            add_log_to("--------------------------------------------------", logger)
            add_log_to("Details:", logger)
            add_log_to(e, logger)
            add_log_to("--------------------------------------------------", logger)
            return True

    def __extract_custom_errors(self, current_connection):
        try:
            cursor = current_connection.cursor()
            status = cursor.var(cx_Oracle.NUMBER)
            line = cursor.var(cx_Oracle.STRING)
            lines = []
            while True:
                cursor.callproc("DBMS_OUTPUT.GET_LINE", (line, status))
                if status.getvalue() == 0:
                    lines.append(line.getvalue())
                else:
                    break
            cursor.close()
            return lines
        except Exception as e:
            print(e)
        return []

    def __get_custom_messgae(self, e: Exception):
        try:
            stringfy = str(e)
            stringfy = stringfy[stringfy.index("::") + 2 :]
            return stringfy[: stringfy.index("::")]
        except Exception as e:
            return "Unspecified error."

    def exec(
        self,
        fn_name: str,
        ret_type: OracleType,
        params: dict = None,
        model: Any = None,
        connection=None,
        custom_params: Optional[List[ZParam]] = None,
        logger: TLogger = None,
    ) -> RepositoryResult:
        """
        DB Funtion executor

        Parameters
        ----------
        fn : str | required
            Function name with package name: PO_LINES_PKG.FN_GET_LINE

        ret_type : OracleType | required
            The return type of oracle db function

        params : Dict | Optional
            Set parameter that the oracle funtion expects

        custom_params : Optional[List[ZParam, IntList, StrList, ClobList]] | Optional
            Custom Set parameter that the oracle funtion expects, see avaialble custom types

        model : marshmallow_objects.models.Model | Optional
            Model specification where the db data will be volcated

        connection : DB Connection | Optional
            The db connection object, if it is not passed by params, it tries to get a global instance

        Raises
        ------
        NotValueProvided
            Connection

        ZDBException
            DB General Exception

        ZIDBException
            Vilated DB Integrity

        Returns
        -------
        result : RespositoryResult
            The result set of oracle db function
        """
        try:
            result = self.execute(
                fn_name=fn_name,
                ret_type=ret_type,
                params=params,
                model=model,
                connection=connection,
                custom_params=custom_params,
                logger=logger,
            )
            return RepositoryResult(data=result, status=ZContractStatus.SUCCESS)
        except ZError as e:
            return RepositoryResult(
                status=ZContractStatus.ERROR, errors=[e.get_error()]
            )

    def execute(
        self,
        fn_name: str,
        ret_type: OracleType,
        params: dict = None,
        model: Any = None,
        connection=None,
        custom_params: Optional[List[ZParam]] = None,
        logger: TLogger = None,
    ):
        """
        New feature for call oracle db functions
        Use this function instead function 'call'

        Parameters
        ----------
        fn : str | required
            Function name with package name: PO_LINES_PKG.FN_GET_LINE

        ret_type : OracleType | required
            The return type of oracle db function

        params : Dict | Optional
            Set parameter that the oracle funtion expects

        custom_params : Optional[List[ZParam, IntList, StrList, ClobList]] | Optional
            Custom Set parameter that the oracle funtion expects, see avaialble custom types

        model : marshmallow_objects.models.Model | Optional
            Model specification where the db data will be volcated

        connection : DB Connection | Optional
            The db connection object, if it is not passed by params, it tries to get a global instance

        Raises
        ------
        NotValueProvided
            Connection

        ZDBException
            DB General Exception

        ZIDBException
            Vilated DB Integrity

        Returns
        -------
        result set : Union[List[Model],int,float,str]
            The result set of oracle db function
        """
        logger = self.logger if logger == None else logger
        if self.db is None:
            add_log_to("DB Connection didn't provided in execute db function", logger)
            raise ZDatabaseOperationError(
                message="Impossible establish a connection with data provider.",
                reason="Connection object did not provided to dispatcher.",
            )
        cn = connection
        connection_passed = True
        if connection == None:
            connection_passed = False
            cn = self.db.get_pool_connection()
        try:
            result_set = self.db.exec(
                fn=fn_name,
                ret_type=ret_type,
                params=params,
                model=model,
                connection=cn,
                custom_params=custom_params,
            )
            if connection_passed == False:
                self.db.release_connection(cn)
            return result_set
        except cx_Oracle.IntegrityError as e:
            if connection_passed == False:
                self.db.release_connection(cn)
            if (
                self.__show_db_error(
                    e,
                    fn_name,
                    "Integrity Database Error",
                    params,
                    custom_params,
                    logger,
                )
                == False
            ):
                add_log_to(e, logger)
            raise ZDBIntegrityError(
                d_except=e,
                message=self.__get_custom_messgae(e),
                reason="A data schema restriction is violated when attempting to process operations on the supplied data.",
            )
        except Exception as e:

            if connection_passed == False:
                self.db.release_connection(cn)
            if (
                self.__show_db_error(
                    e,
                    fn_name,
                    "Unspecified DB Exception",
                    params,
                    custom_params,
                    logger,
                )
                == False
            ):
                add_log_to(e, logger)

            raise ZDatabaseOperationError(
                d_except=e,
                message=self.__get_custom_messgae(e),
                reason="When trying to process an operation with the database provider.",
            )

    def exec_txn(
        self,
        fn_name: str,
        ret_type: OracleType,
        table_name: str = None,
        operation_type=ZOperation,
        transaction: str = None,
        params: dict = None,
        model: Any = None,
        connection=None,
        custom_params: Optional[List[ZParam]] = None,
        logger: TLogger = None,
    ) -> RepositoryResult:
        self._check_transact()
        result = self.exec(
            fn_name=fn_name,
            ret_type=ret_type,
            params=params,
            model=model,
            connection=connection,
            custom_params=custom_params,
            logger=logger,
        )
        if result.status == ZContractStatus.SUCCESS:
            row = result
            if "INT" in operation_type.value:
                row = params
            else:
                if ret_type == OracleType.number or ret_type == OracleType.decimal:
                    row = {"id": result}
            self.transact.operation(operation_type, table_name, row, transaction)
        return result

    def execute_txn(
        self,
        fn_name: str,
        ret_type: OracleType,
        table_name: str = None,
        operation_type=ZOperation,
        transaction: str = None,
        params: dict = None,
        model: Any = None,
        connection=None,
        custom_params: Optional[List[ZParam]] = None,
        logger: TLogger = None,
    ) -> Any:
        """
        Can be throw exception
        """
        self._check_transact()
        result = self.execute(
            fn_name=fn_name,
            ret_type=ret_type,
            params=params,
            model=model,
            connection=connection,
            custom_params=custom_params,
            logger=logger,
        )

        if result != None:
            row = result
            if "INT" in operation_type.value:
                row = params
            else:
                if ret_type == OracleType.number or ret_type == OracleType.decimal:
                    row = {"id": result}
            self.transact.operation(operation_type, table_name, row, transaction)
        return result

    def insert(
        self,
        fn_name: str,
        ret_type: OracleType,
        table_name: str = None,
        transaction: str = None,
        params: dict = None,
        model: Any = None,
        connection=None,
        custom_params: Optional[List[ZParam]] = None,
        logger: TLogger = None,
    ) -> Any:
        """
        Zurck'z
        Can be throw an exception
        """
        self._check_transact()
        result = self.execute(
            fn_name=fn_name,
            ret_type=ret_type,
            params=params,
            model=model,
            connection=connection,
            custom_params=custom_params,
            logger=logger,
        )

        if result != None:
            row = result
            if ret_type == OracleType.number or ret_type == OracleType.decimal:
                row = {"id": result}
            self.transact.insert(table=table_name, data=row, transaction=transaction)
        return result

    def int_insert(
        self,
        fn_name: str,
        ret_type: OracleType,
        table_name: str = None,
        transaction: str = None,
        params: dict = None,
        model: Any = None,
        connection=None,
        custom_params: Optional[List[ZParam]] = None,
        logger: TLogger = None,
    ) -> Any:
        """
        Zurck'z
        Can be throw an exception
        """
        self._check_transact()
        result = self.execute(
            fn_name=fn_name,
            ret_type=ret_type,
            params=params,
            model=model,
            connection=connection,
            custom_params=custom_params,
            logger=logger,
        )

        if result != None:
            self.transact.int_insert(
                table=table_name, data=params, transaction=transaction
            )
        return result

    def update(
        self,
        fn_name: str,
        ret_type: OracleType,
        table_name: str = None,
        transaction: str = None,
        params: dict = None,
        model: Any = None,
        connection=None,
        custom_params: Optional[List[ZParam]] = None,
        logger: TLogger = None,
    ) -> Any:
        """
        Zurck'z
        Can be throw an exception
        """
        self._check_transact()
        result = self.execute(
            fn_name=fn_name,
            ret_type=ret_type,
            params=params,
            model=model,
            connection=connection,
            custom_params=custom_params,
            logger=logger,
        )

        if result != None:
            row = result
            if ret_type == OracleType.number or ret_type == OracleType.decimal:
                row = {"id": result}
            self.transact.update(table=table_name, data=row, transaction=transaction)
        return result

    def int_update(
        self,
        fn_name: str,
        ret_type: OracleType,
        table_name: str = None,
        transaction: str = None,
        params: dict = None,
        model: Any = None,
        connection=None,
        custom_params: Optional[List[ZParam]] = None,
        logger: TLogger = None,
    ) -> Any:
        """
        Zurck'z
        Can be throw an exception
        """
        self._check_transact()
        result = self.execute(
            fn_name=fn_name,
            ret_type=ret_type,
            params=params,
            model=model,
            connection=connection,
            custom_params=custom_params,
            logger=logger,
        )

        if result != None:
            self.transact.int_update(
                table=table_name, data=params, transaction=transaction
            )
        return result

    def delete(
        self,
        fn_name: str,
        ret_type: OracleType,
        table_name: str = None,
        transaction: str = None,
        params: dict = None,
        model: Any = None,
        connection=None,
        custom_params: Optional[List[ZParam]] = None,
        logger: TLogger = None,
    ) -> Any:
        """
        Zurck'z
        Can be throw an exception
        """
        self._check_transact()
        result = self.execute(
            fn_name=fn_name,
            ret_type=ret_type,
            params=params,
            model=model,
            connection=connection,
            custom_params=custom_params,
            logger=logger,
        )

        if result != None:
            row = result
            if ret_type == OracleType.number or ret_type == OracleType.decimal:
                row = {"id": result}
            self.transact.delete(table=table_name, data=row, transaction=transaction)
        return result

    def int_delete(
        self,
        fn_name: str,
        ret_type: OracleType,
        table_name: str = None,
        transaction: str = None,
        params: dict = None,
        model: Any = None,
        connection=None,
        custom_params: Optional[List[ZParam]] = None,
        logger: TLogger = None,
    ) -> Any:
        """
        Zurck'z
        Can be throw an exception
        """
        self._check_transact()
        result = self.execute(
            fn_name=fn_name,
            ret_type=ret_type,
            params=params,
            model=model,
            connection=connection,
            custom_params=custom_params,
            logger=logger,
        )

        if result != None:
            self.transact.int_delete(
                table=table_name, data=params, transaction=transaction
            )
        return result

    def begin_txn(self, name: str):
        """
        """
        self._check_transact()
        self.transact.begin_txn(name)

    def commit_txn(self, name: str = None, pop: bool = False):
        """
        """
        self._check_transact()
        return self.transact.commit(name, pop)

    def get_transaction_store(self) -> dict:
        self._check_transact()
        return self.transact.store


__author__ = "Noé Cruz | contactozurckz@gmail.com"
__copyright__ = "Copyright 2021, Small APi Project"
__credits__ = ["Noé Cruz", "Zurck'z"]
__license__ = "MIT"
__version__ = "0.0.1"
__maintainer__ = "Noé Cruz"
__email__ = "contactozurckz@gmail.com"
__status__ = "Dev"


from zpy.api.repository import RepositoryResult
from typing import Any, List, TypeVar
from zpy.api import ZContract, ZContractStatus
from zpy.api.reponse import Status

T = TypeVar("T")


class ServiceResult(ZContract[T]):
    http_status: Status

    def __init__(
        self,
        data: T = None,
        status: ZContractStatus = ZContractStatus.PENDING ,
        errors: List[Any] = None,
        http_status: Status = Status.SUCCESS,
    ) -> None:
        super().__init__(data=data, status=status, errors=errors)
        self.http_status = http_status


class ZService:
    def __init__(self) -> None:
        print(f"Zurck'z  - Core Service was iniitalized")

    def either_error(self, errors: List[dict], result: RepositoryResult) -> T:
        if result.status == ZContractStatus.ERROR:
            errors.extend(result.errors)
        return result.data


from enum import Enum


class ZOperation(Enum):
    INSERT = "INSERT"  # Insert that return nre id
    UPDATE = "UPDATE"  # Update that return row before updated
    DELETE = "DELETE"  # Delete return row deleted
    INT_INSERT = "IN_INSERT"  # Insert that return row with data inserted
    INT_UPDATE = "IN_UPDATE"  # Update that return row with data inserted
    INT_DELETE = "IN_DELETE"  # Delete return row with data inserted


class ZTranstaction(object):

    default_transaction: str = "default"
    store = None

    def __init__(self) -> None:
        super().__init__()
        self.store = {self.default_transaction: {}}

    def begin_txn(self, name: str):
        self.store[name] = {}

    def operation(
        self, operation: ZOperation, table: str, data: dict, transaction: str = None
    ):
        transact = transaction or self.default_transaction
        operation = {"type": operation.value, "data": data}
        if table in self.store[transact]:
            self.store[transact][table].append(operation)
        else:
            self.store[transact][table] = [operation]

    def insert(self, table: str, data: dict, transaction: str = None):
        self.operation(ZOperation.INSERT, table, data, transaction)

    def update(self, table: str, data: dict, transaction: str = None):
        self.operation(ZOperation.UPDATE, table, data, transaction)

    def delete(self, table: str, data: dict, transaction: str = None):
        self.operation(ZOperation.DELETE, table, data, transaction)

    def int_insert(self, table: str, data: dict, transaction: str = None):
        self.operation(ZOperation.INT_INSERT, table, data, transaction)

    def int_update(self, table: str, data: dict, transaction: str = None):
        self.operation(ZOperation.INT_UPDATE, table, data, transaction)

    def int_delete(self, table: str, data: dict, transaction: str = None):
        self.operation(ZOperation.INT_DELETE, table, data, transaction)

    def commit(self, name: str = None, pop: bool = False):
        print(self.store)
        transact = name
        if name not in self.store or name == None:
            transact = self.default_transaction
        if pop:
            return self.store.pop(transact, None)
        return self.store[transact]


from zpy.utils import get_operation_id
from enum import Enum
from typing import Any
import json


__author__ = "Noé Cruz | contactozurckz@gmail.com"
__copyright__ = "Copyright 2021, Small APi Project"
__credits__ = ["Noé Cruz", "Zurck'z"]
__license__ = "MIT"
__version__ = "0.0.1"
__maintainer__ = "Noé Cruz"
__email__ = "contactozurckz@gmail.com"
__status__ = "Dev"


PAYLOAD_KEY: str = "data"
STATUS_KEY: str = "status"
DESCRIPTION_KEY: str = "description"
OPERATION_ID_KEY: str = "operationId"
ERRORS_KEY: str = "errors"


class Status(Enum):
    """
    Common HTTP status codes

    CODE | SHORT DESCRIPTION | STATUS DETAILS
    """

    SUCCESS                = (200, "SUCCEEDED", "The request has succeeded")
    CREATED                = (201, "CREATED","The request has been fulfilled and resulted in a new resource being created.")
    ACCEPTED               = (202, "ACCEPTED","The request has been accepted for processing, but the processing has not been completed.")
    NO_CONTENT             = (204, "NO CONTENT","The request has been completed successfully but your response has no content, although the headers can be useful.",)
    PARTIAL_CONTENT        = (206, "PARTIAL CONTENT", "Partial content")

    BAD_REQUEST            = (400, "BAD REQUEST","The request could not be understood by the server due to malformed syntax. The client SHOULD NOT repeat the request without modifications.",)
    UNAUTHORIZED           = (401, "UNAUTHORIZED", "The request requires user authentication.")
    FORBIDDEN              = (403, "FORBIDDEN","The server understood the request, but is refusing to fulfill it.")
    NOT_FOUND              = (404, "NOT FOUND","The server has not found anything matching the Request-URI.",)
    METHOD_NOT_ALLOWED     = (405, "METHOD NOT ALLOWED","The method specified in the Request-Line is not allowed for the resource identified by the Request-URI.",)
    CONTENT_NOT_ACCEPTABLE = (406, "METHOD NOT ACCEPTABLE","The resource identified by the request is only capable of generating response entities which have content characteristics not acceptable according to the accept headers sent in the request.",)
    REQUEST_TIMEOUT        = (408, "REQUEST TIMEOUT", "Time out")
    PRE_CONDITION_FAILED   = (412, "PRECONDITION FAILED","The client has indicated preconditions in their headers which the server does not meet.",)
    UNSUPPORTED_MEDIA_TYPE = (415, "UNSUPPORTED MEDIA TYPE","The multimedia format of the requested data is not supported by the server, therefore the server rejects the request.",)
    IM_A_TEAPOT            = (418, "IM A TEAPOT","The server refuses to try to make coffee with a kettle.",)
    CONFLICT               = (409, "CONFLICT", "The server found conflict with request supplied.")
    UNPROCESSABLE          = (422, "UNPROCESSABLE ENTITY","The process could not be completed due to a semantics error.",)
    LOCKED                 = (423, "LOCKED","The source or destination resource of a method is locked.",)

    INTERNAL_SERVER_ERROR  = (500, "INTERNAL SERVER ERROR","The server encountered an unexpected condition which prevented it from fulfilling the request.",)
    NOT_IMPEMENTED         = (501, "NOT IMPLEMENTED","The server does not support the functionality required to fulfill the request",)
    SERVICE_UNAVAIBLE      = (503, "SERVICE UNAVAIBLE","The server is currently unable to handle the request due to a temporary overloading or maintenance of the server.",)
    GATEWAY_TIMEOUT        = (503, "GATEWAY TIMEOUT", "Timeout")
    LOOP_DETECTED          = (508, "LOOP DETECTED","The server encountered an infinite loop while processing the request. ",)


class Error:
    pass


class Builder:
    """
    Common resopnse builder
    """

    response = {
        PAYLOAD_KEY: {},
        DESCRIPTION_KEY: Status.SUCCESS.value[2],
        STATUS_KEY: Status.SUCCESS.value[1],
        OPERATION_ID_KEY: None,
    }
    response_code = 200

    def __init__(self, custom_schema=None) -> None:
        if custom_schema is not None:
            self.response = custom_schema

    def add_custom(self, key: str, data: Any):
        """Add custom item to object body.

        PD: Replacemnt item if response body have a item

        Parameters
        ----------
        key : str | required
            The item key

        data : Any | required
            The item value

        Raises
        ------
        NotValueProvided
            If no values is set for the response or passed in as a
            parameter.
        """
        self.response[key] = data
        return self

    def add_error(self, error: any):
        """Add error to erros object body.

        PD: Replacemnt item if response body have a item

        Parameters
        ----------
        error : Any, str, dict | required
            The item error response

        Raises
        ------
        NotErrorProvided
            If no status is set for the response or passed in as a
            parameter.
        """
        if ERRORS_KEY in self.response:
            self.response[ERRORS_KEY].append(error)
        else:
            self.response[ERRORS_KEY] = [error]
        return self

    def add_payload(self, payload):
        """Add data payload to reponse body.

        PD: Replacemnt item if response body have a item

        Parameters
        ----------
        payload : Any, str, dict | required
            The data response

        Raises
        ------
        NotPayloadProvided
            If no status is set for the response or passed in as a
            parameter.
        """
        self.response[PAYLOAD_KEY] = payload
        return self

    def add_description(self, description: str):
        """Add description to reponse body.

        PD: Replacemnt if response body have a description

        Parameters
        ----------
        description : str | required
            The description response

        Raises
        ------
        NotDescriptionProvided
            If no status is set for the response or passed in as a
            parameter.
        """
        self.response[DESCRIPTION_KEY] = description
        return self

    def add_operation_id(self, id: str):
        """Add operation to reponse body.

        PD: Replacemnt if response body have a item

        Parameters
        ----------
        id : str | required
            The operation id response

        Raises
        ------
        NotValueProvided
            If no status is set for the response or passed in as a
            parameter.
        """
        self.response[OPERATION_ID_KEY] = id
        return self

    def add_status(self, status: Status):
        """Add status to reponse body.

        The status value is a tuple conformed by 3 parts:

        CODE | SHORT DESCRIPTION | STATUS DETAILS

        Parameters
        ----------
        status : Status | Tuple, required
            The response information

        Raises
        ------
        NotStatusProvided
            If no status is set for the response or passed in as a
            parameter.
        """
        self.response[STATUS_KEY] = status.value[1]
        self.response_code = status.value[0]
        self.response[DESCRIPTION_KEY] = status.value[2]
        return self

    @staticmethod
    def success(
        payload={},
        status_code: Status = Status.SUCCESS,
        operation_id=None,
        description: str = None,
        fn_transform = None
    ):
        response = {}
        response[OPERATION_ID_KEY] = (
            get_operation_id() if operation_id is None else operation_id
        )
        response[PAYLOAD_KEY] = payload
        response[DESCRIPTION_KEY] = (
            status_code.value[2] if description is None else description
        )
        response[STATUS_KEY] = status_code.value[1]
        if fn_transform != None:
            response = fn_transform(response)
        return response, status_code.value[0]

    @staticmethod
    def error(
        payload={},
        errors=[],
        status=Status.INTERNAL_SERVER_ERROR,
        operation_id=None,
        description: str = None,
        fn_transform=None,
    ):
        response = {}
        response[OPERATION_ID_KEY] = (
            get_operation_id() if operation_id is None else operation_id
        )
        response[ERRORS_KEY] = errors
        response[PAYLOAD_KEY] = payload
        response[DESCRIPTION_KEY] = (
            status.value[2] if description is None else description
        )
        response[STATUS_KEY] = status.value[1]

        if fn_transform != None:
            response = fn_transform(response)

        return response, status.value[0]

    @staticmethod
    def bad_request(
        payload={},
        errors=[],
        status=Status.BAD_REQUEST,
        operation_id=None,
        description: str = None,
        fn_transform=None,
    ):
        response = {}
        response[OPERATION_ID_KEY] = (
            get_operation_id() if operation_id is None else operation_id
        )
        response[ERRORS_KEY] = errors
        response[PAYLOAD_KEY] = payload
        response[DESCRIPTION_KEY] = (
            status.value[2] if description is None else description
        )
        response[STATUS_KEY] = status.value[1]
        if fn_transform != None:
            response = fn_transform(response)
        return response, status.value[0]

    @staticmethod
    def no_content(
        payload={},
        errors=[],
        status=Status.NO_CONTENT,
        operation_id=None,
        description: str = None,
        fn_transform=None,
    ):
        response = {}
        response[OPERATION_ID_KEY] = (
            get_operation_id() if operation_id is None else operation_id
        )
        response[ERRORS_KEY] = errors
        response[PAYLOAD_KEY] = payload
        response[DESCRIPTION_KEY] = (
            status.value[2] if description is None else description
        )
        response[STATUS_KEY] = status.value[1]
        if fn_transform != None:
            response = fn_transform(response)
        return response, status.value[0]

    def build(self,fn_transform=None) -> Any:
        if self.response[OPERATION_ID_KEY] is None:
            self.response[OPERATION_ID_KEY] = get_operation_id()
        if fn_transform != None:
            self.response = fn_transform(self.response)
        return self.response, self.response_code

    def str_json_build(self,fn_transform) -> Any:
        if self.response[OPERATION_ID_KEY] is None:
            self.response[OPERATION_ID_KEY] = get_operation_id()
        if fn_transform != None:
            self.response = fn_transform(self.response)
        return json.dumps(self.response), self.response_code


class BResponse(Builder):
    """
    Zurck'z Common Builder Response

    """

    def err(
        payload=None,
        errors=[],
        status=Status.INTERNAL_SERVER_ERROR,
        operation_id=None,
        description: str = None,
        fn_transform=None,
    ):
        """
        Build a error response with None payload
        """
        return Builder.error(
            payload, errors, status, operation_id, description, fn_transform
        )



from abc import abstractmethod

__author__ = "Noé Cruz | contactozurckz@gmail.com"
__copyright__ = "Copyright 2021, Small APi Project"
__credits__ = ["Noé Cruz", "Zurck'z"]
__license__ = "upax"
__version__ = "0.0.1"
__maintainer__ = "Noé Cruz"
__email__ = "contactozurckz@gmail.com"
__status__ = "Dev"


class Stage():
    name = ""
    description = ""    
    def __init__(self,name,description) -> None:
        self.name = name
        self.description = description

    @abstractmethod
    def get(self) -> str:
        return "\n[ PH: :: {0} ]\n[ PD: :: {1} ]\n".format(self.name,self.description)

    @abstractmethod
    def desc(self) -> str:
        return "[ PD: :: {0} ]".format(self.description)

    @abstractmethod
    def title(self) -> str:
        return "[ PH: :: {0} ]".format(self.name)

class Unspecified(Stage):
    name = "Unspecified Stage"
    description = "Unspecified stage by developer"
    def __init__(self,name: str=None,description: str=None) -> None:
        super().__init__(
            name=self.name if name is None else name,
            description=self.description if description is None else description
        )

class StartingCoreModule(Stage):
    name = "Starting Core Module"
    description = "Initializing core modules"
    def __init__(self,name: str=None,description: str=None) -> None:
        super().__init__(
            name=self.name if name is None else name,
            description=self.description if description is None else description
        )

class ResourceGateway(Stage):
    name = "Resource Gateway"
    description = "HTTP Resource Gateway"
    def __init__(self,name: str=None,description: str=None) -> None:
        super().__init__(
            name=self.name if name is None else name,
            description=self.description if description is None else description
        )
class Decrypt(Stage):
    name = "Decrypt Stage"
    description = "Http request decryption Stage"

    def __init__(self,name: str=None,description: str=None) -> None:
        super().__init__(
            name=self.name if name is None else name,
            description=self.description if description is None else description
        )

class SchemaValidation(Stage):
    name = "Schema Validation Stage"
    description = "Schemes or request models are validated according to the validations given"
    def __init__(self,name: str=None,description: str=None) -> None:
        super().__init__(
            name=self.name if name is None else name,
            description=self.description if description is None else description
        )

class ResourceLayer(Stage):
    name = "Resource Layer Stage"
    description = "Abstract presentation to business logic."
    def __init__(self,name: str=None,description: str=None) -> None:
        super().__init__(
            name=self.name if name is None else name,
            description=self.description if description is None else description
        )

class BusinessLogicExecution(Stage):
    name = "Business Logic Layer Stage"
    description = "Business Logic execution"
    def __init__(self,name: str=None,description: str=None) -> None:
        super().__init__(
            name=self.name if name is None else name,
            description=self.description if description is None else description
        )


class AuthenticationCheck(Stage):
    name = "Authentication Check Layer Stage"
    description = "Authentication Check"
    def __init__(self,name: str=None,description: str=None) -> None:
        super().__init__(
            name=self.name if name is None else name,
            description=self.description if description is None else description
        )

class AuthorizationCheck(Stage):
    name = "Authorization Check Layer Stage"
    description = "Authorization Check"
    def __init__(self,name: str=None,description: str=None) -> None:
        super().__init__(
            name=self.name if name is None else name,
            description=self.description if description is None else description
        )

class ServiceLayer(Stage):
    name = "Service Layer Stage"
    description = "Service Layer, all buisiness logic execute"
    def __init__(self,name: str=None,description: str=None) -> None:
        super().__init__(
            name=self.name if name is None else name,
            description=self.description if description is None else description
        )

class DataAccess(Stage):
    name = "Data Access Stage"
    description = "Data Access"
    def __init__(self,name: str=None,description: str=None) -> None:
        super().__init__(
            name=self.name if name is None else name,
            description=self.description if description is None else description
        )

class BuildingReponses(Stage):
    name = "Building Response Stage"
    description = "Building response..."
    def __init__(self,name: str=None,description: str=None) -> None:
        super().__init__(
            name=self.name if name is None else name,
            description=self.description if description is None else description
        )

class Encryption(Stage):
    name = "Encryption Stage"
    description = "Encryption Stage", "Response encryption proccess"
    def __init__(self,name: str=None,description: str=None) -> None:
        super().__init__(
            name=self.name if name is None else name,
            description=self.description if description is None else description
        )    
